# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Optional, Any, Union

from .context.context import ContextContainer
from .db import IconScoreDatabase, Key, KeyType
from ..base.address import Address
from ..base.exception import InvalidParamsException, InvalidContainerAccessException
from ..icon_constant import IconScoreContextType, Revision
from ..utils import int_to_bytes, bytes_to_int

K = TypeVar('K', int, str, Address, bytes)
V = TypeVar('V', int, str, Address, bytes, bool)


def get_encoded_key(key: V) -> bytes:
    return ContainerUtil.encode_key(key)


class ContainerUtil(object):

    @classmethod
    def encode_key(cls, key: K) -> bytes:
        """Create a key passed to IconScoreDatabase

        :param key:
        :return:
        """
        if key is None:
            raise InvalidParamsException('key is None')

        if isinstance(key, int):
            bytes_key = int_to_bytes(key)
        elif isinstance(key, str):
            bytes_key = key.encode('utf-8')
        elif isinstance(key, Address):
            bytes_key = key.to_bytes()
        elif isinstance(key, bytes):
            bytes_key = key
        else:
            raise InvalidParamsException(f'Unsupported key type: {type(key)}')
        return bytes_key

    @classmethod
    def encode_value(cls, value: V) -> bytes:
        if isinstance(value, int):
            byte_value = int_to_bytes(value)
        elif isinstance(value, str):
            byte_value = value.encode('utf-8')
        elif isinstance(value, Address):
            byte_value = value.to_bytes()
        elif isinstance(value, bool):
            byte_value = int_to_bytes(int(value))
        elif isinstance(value, bytes):
            byte_value = value
        else:
            raise InvalidParamsException(f'Unsupported value type: {type(value)}')
        return byte_value

    @classmethod
    def decode_object(cls, value: bytes, value_type: type) -> Optional[Union[K, V]]:
        if value is None:
            return get_default_value(value_type)

        obj_value = None
        if value_type == int:
            obj_value = bytes_to_int(value)
        elif value_type == str:
            obj_value = value.decode()
        elif value_type == Address:
            obj_value = Address.from_bytes(value)
        if value_type == bool:
            obj_value = bool(bytes_to_int(value))
        elif value_type == bytes:
            obj_value = value
        return obj_value


class DictDB(object):
    """
    Utility classes wrapping the state DB.
    DictDB behaves more like python dict.
    DictDB does not maintain order.

    :K: [int, str, Address, bytes]
    :V: [int, str, Address, bytes, bool]
    """

    def __init__(self,
                 var_key: K,
                 db: IconScoreDatabase,
                 value_type: type,
                 depth: int = 1) -> None:
        if not (1 <= depth <= 5):
            raise InvalidParamsException(f"Depth out of range: {depth}")

        self._db = db.get_sub_db(Key(get_encoded_key(var_key), KeyType.DICT))
        self.__value_type = value_type
        self.__depth = depth

    def remove(self, key: K) -> None:
        """
        Removes the value of given key

        :param key:
        """
        if self.__depth != 1:
            raise InvalidContainerAccessException('DictDB depth mismatch')
        self._db.delete(get_encoded_key(key))

    def __setitem__(self, key: K, value: V) -> None:
        if self.__depth != 1:
            raise InvalidContainerAccessException('DictDB depth mismatch')

        encoded_key: bytes = get_encoded_key(key)
        encoded_value: bytes = ContainerUtil.encode_value(value)

        self._db.put(encoded_key, encoded_value)

    def __getitem__(self, key: K) -> Any:
        if self.__depth == 1:
            encoded_key: bytes = get_encoded_key(key)
            return ContainerUtil.decode_object(self._db.get(encoded_key), self.__value_type)
        else:
            return DictDB(key, self._db, self.__value_type, self.__depth - 1)

    def __delitem__(self, key: K):
        self.remove(key)

    def __contains__(self, key: K):
        # Plyvel doesn't allow setting None value in the DB.
        # so there is no case of returning None value if the key exists.
        value = self._db.get(get_encoded_key(key))
        return value is not None

    def __iter__(self):
        raise InvalidContainerAccessException("Iteration not supported in DictDB")


class ArrayDB(object):
    """
    Utility classes wrapping the state DB.
    ArrayDB supports length and iterator, maintains order.

    :K: [int, str, Address, bytes]
    :V: [int, str, Address, bytes, bool]
    """
    SIZE_BYTE_KEY = b"size"

    def __init__(self, var_key: K, db: 'IconScoreDatabase', value_type: type) -> None:
        self._db: IconScoreDatabase = db.get_sub_db(Key(get_encoded_key(var_key), KeyType.ARRAY))

        self.__value_type = value_type
        self.__legacy_size = self.__get_size_from_db()

    def put(self, value: V) -> None:
        """
        Puts the value at the end of array

        :param value: value to add
        """
        size: int = self.__get_size()
        self.__put(size, value)
        self.__set_size(size + 1)

    def pop(self) -> Optional[V]:
        """
        Gets and removes last added value

        :return: last added value
        """
        size: int = self.__get_size()
        if size == 0:
            return None

        index = size - 1
        last_val = self[index]
        self._db.delete(get_encoded_key(index))
        self.__set_size(index)
        return last_val

    def get(self, index: int = 0) -> V:
        """
        Gets the value at index

        :param index: index
        :return: value at the index
        """
        return self.__getitem__(index)

    def __get_size(self) -> int:
        if self.__is_defective_revision():
            return self.__legacy_size
        else:
            return self.__get_size_from_db()

    def __get_size_from_db(self) -> int:
        value: bytes = self._db.get(Key(self.SIZE_BYTE_KEY, KeyType.ARRAY_SIZE))
        return ContainerUtil.decode_object(value, int)

    def __set_size(self, size: int) -> None:
        self.__legacy_size = size
        byte_value = ContainerUtil.encode_value(size)

        key = Key(self.SIZE_BYTE_KEY, KeyType.ARRAY_SIZE)
        self._db.put(key, byte_value)

    def __put(self, index: int, value: V) -> None:
        byte_value = ContainerUtil.encode_value(value)
        self._db.put(get_encoded_key(index), byte_value)

    def __iter__(self):
        size: int = self.__get_size()

        for i in range(size):
            key: bytes = get_encoded_key(i)
            value: bytes = self._db.get(key)
            yield ContainerUtil.decode_object(value, self.__value_type)

    def __len__(self):
        return self.__get_size()

    def __setitem__(self, index: int, value: V) -> None:
        if not isinstance(index, int):
            raise InvalidParamsException('Invalid index type: not an integer')

        size: int = self.__get_size()
        index: int = self._to_positive_index(index, size)
        self.__put(index, value)

    def __getitem__(self, index: int) -> V:
        if not isinstance(index, int):
            raise InvalidParamsException('Invalid index type: not an integer')

        size: int = self.__get_size()
        index: int = self._to_positive_index(index, size)

        key: bytes = get_encoded_key(index)
        value: bytes = self._db.get(key)
        return ContainerUtil.decode_object(value, self.__value_type)

    def __contains__(self, item: V):
        for e in self:
            if e == item:
                return True
        return False

    @classmethod
    def __is_defective_revision(cls):
        context = ContextContainer._get_context()
        revision = context.revision
        return context.type == IconScoreContextType.INVOKE and revision < Revision.THREE.value

    @classmethod
    def _to_positive_index(cls, index: int, size: int) -> int:
        if index < 0:
            index += size

        if 0 <= index < size:
            return index

        raise InvalidParamsException('ArrayDB out of index')


class VarDB(object):
    """
    Utility classes wrapping the state DB.
    VarDB can be used to store simple key-value state.

    :K: [int, str, Address, bytes]
    :V: [int, str, Address, bytes, bool]
    """
    _DUMMY_LAST_KEY = Key(b"", KeyType.DUMMY)

    def __init__(self, var_key: K, db: 'IconScoreDatabase', value_type: type) -> None:
        # Use var_key as a db prefix in the case of VarDB
        self._db = db.get_sub_db(Key(get_encoded_key(var_key), KeyType.VAR))
        self.__value_type = value_type

    def set(self, value: V) -> None:
        """
        Sets the value

        :param value: a value to be set
        """
        byte_value = ContainerUtil.encode_value(value)
        self._db.put(self._DUMMY_LAST_KEY, byte_value)

    def get(self) -> Optional[V]:
        """
        Gets the value

        :return: value of the var db
        """
        return ContainerUtil.decode_object(
            self._db.get(self._DUMMY_LAST_KEY),
            self.__value_type
        )

    def remove(self) -> None:
        """
        Deletes the value
        """
        self._db.delete(self._DUMMY_LAST_KEY)


def get_default_value(value_type: type) -> Any:
    if value_type == int:
        return 0
    elif value_type == str:
        return ""
    elif value_type == bool:
        return False
    return None



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
from collections import namedtuple
from typing import TYPE_CHECKING, Optional, Tuple, List, Set

from iconcommons import Logger
from ..reward_calc.msg_data import Header, TxData, PRepsData, TxType, make_block_produce_info_key
from ...base.exception import DatabaseException, InternalServiceErrorException
from ...database.db import KeyValueDatabase
from ...icon_constant import (
    DATA_BYTE_ORDER, Revision, RC_DATA_VERSION_TABLE, RC_DB_VERSION_0,
    IISS_LOG_TAG, ROLLBACK_LOG_TAG
)
from ...iiss.reward_calc.data_creator import DataCreator
from ...utils import bytes_to_hex
from ...utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ...base.address import Address
    from ...database.wal import IissWAL
    from ..reward_calc.msg_data import Data, DelegationInfo
    from ...iconscore.icon_score_context import IconScoreContext


RewardCalcDBInfo = namedtuple('RewardCalcDBInfo', ['path', 'block_height'])


def get_rc_version(revision: int) -> int:
    while revision >= Revision.IISS.value:
        version: int = RC_DATA_VERSION_TABLE.get(revision, -1)
        if version > -1:
            return version

        revision -= 1

    return RC_DB_VERSION_0


def get_version_and_revision(db: 'KeyValueDatabase') -> Tuple[int, int]:
    version_and_revision: Optional[bytes] = db.get(Storage.KEY_FOR_VERSION_AND_REVISION)
    if version_and_revision is None:
        return -1, -1

    version, revision = MsgPackForDB.loads(version_and_revision)
    return version, revision


class Storage(object):
    """Manages RC DB which Reward Calculator will use to calculate a reward for each address

    """

    CURRENT_IISS_DB_NAME = "current_db"
    STANDBY_IISS_DB_NAME_PREFIX = "standby_rc_db"
    IISS_RC_DB_NAME_PREFIX = "iiss_rc_db"

    KEY_FOR_GETTING_LAST_TRANSACTION_INDEX = b'last_transaction_index'
    KEY_FOR_CALC_RESPONSE_FROM_RC = b'calc_response_from_rc'
    KEY_FOR_VERSION_AND_REVISION = b'version_and_revision'

    def __init__(self):
        self._path: str = ""
        self._db: Optional['KeyValueDatabase'] = None
        # 'None' if open() is not called else 'int'
        self._db_iiss_tx_index: int = -1

    def open(self, context: 'IconScoreContext', path: str):
        revision: int = context.revision

        if not os.path.exists(path):
            raise DatabaseException(f"Invalid IISS DB path: {path}")
        self._path = path

        self._db = self.create_current_db(path)

        self._db_iiss_tx_index = self._load_last_transaction_index()
        Logger.info(tag=IISS_LOG_TAG, msg=f"last_transaction_index on open={self._db_iiss_tx_index}")

        # todo: check side effect of WAL
        self._supplement_db(context, revision)

    def rollback(self, _context: 'IconScoreContext', block_height: int, block_hash: bytes):
        Logger.info(tag=ROLLBACK_LOG_TAG,
                    msg=f"rollback() start: block_height={block_height} block_hash={bytes_to_hex(block_hash)}")

        if self._db is not None:
            raise InternalServiceErrorException("current_db has been opened on rollback")

        if not os.path.exists(self._path):
            raise DatabaseException(f"Invalid IISS DB path: {self._path}")

        self._db = self.create_current_db(self._path)

        self._db_iiss_tx_index = self._load_last_transaction_index()
        Logger.info(tag=IISS_LOG_TAG, msg=f"last_transaction_index on open={self._db_iiss_tx_index}")

        Logger.info(tag=ROLLBACK_LOG_TAG, msg="rollback() end")

    @property
    def key_value_db(self) -> 'KeyValueDatabase':
        return self._db

    def _supplement_db(self, context: 'IconScoreContext', revision: int):
        # Supplement db which is made by previous icon service version (as there is no version, revision and header)
        if revision < Revision.IISS.value:
            return

        rc_version, _ = self.get_version_and_revision()
        if rc_version == -1:
            self._put_version_and_revision(revision)

        # On the first change point.
        # We have to put Header for RC
        if self._db.get(Header.PREFIX) is None:
            rc_version, rc_revision = self.get_version_and_revision()
            end_block_height: int = context.storage.iiss.get_end_block_height_of_calc(context)
            calc_period: int = context.storage.iiss.get_calc_period(context)
            prev_end_calc_block_height: int = end_block_height - calc_period

            # if this point is new calc start point ...
            # we have to set block height in header data.
            if prev_end_calc_block_height == context.block.height:
                end_block_height: int = context.block.height
            header: 'Header' = DataCreator.create_header(rc_version, end_block_height, rc_revision)
            self.put_data_directly(header)

            Logger.debug(tag=IISS_LOG_TAG, msg=f"No header data. Put Header to db on open: {str(header)}")

    @classmethod
    def get_standby_rc_db_name(cls, block_height: int) -> str:
        return cls._get_db_name(cls.STANDBY_IISS_DB_NAME_PREFIX, block_height)

    @classmethod
    def get_iiss_rc_db_name(cls, block_height: int) -> str:
        return cls._get_db_name(cls.IISS_RC_DB_NAME_PREFIX, block_height)

    @classmethod
    def _get_db_name(cls, prefix: str, block_height: int) -> str:
        return f"{prefix}_{block_height}"

    def put_data_directly(self, iiss_data: 'Data', tx_index: Optional[int] = None):
        if isinstance(iiss_data, TxData):
            key: bytes = iiss_data.make_key(tx_index)
            value: bytes = iiss_data.make_value()
        else:
            key: bytes = iiss_data.make_key()
            value: bytes = iiss_data.make_value()
        self._db.put(key, value)

    def close(self):
        """Close the embedded database.
        """
        if self._db:
            self._db.close()
            self._db = None

    def put_calc_response_from_rc(self, iscore: int, block_height: int, state_hash: bytes):
        version = 1
        response_from_rc: bytes = MsgPackForDB.dumps([version, iscore, block_height, state_hash])
        self._db.put(self.KEY_FOR_CALC_RESPONSE_FROM_RC, response_from_rc)

    def get_calc_response_from_rc(self) -> Tuple[int, int, Optional[bytes]]:
        response_from_rc: Optional[bytes] = self._db.get(self.KEY_FOR_CALC_RESPONSE_FROM_RC)
        if response_from_rc is None:
            return -1, -1, None
        response_from_rc: list = MsgPackForDB.loads(response_from_rc)
        version = response_from_rc[0]
        if version == 0:
            iscore = response_from_rc[1]
            block_height = response_from_rc[2]
            state_hash = None
        elif version == 1:
            iscore = response_from_rc[1]
            block_height = response_from_rc[2]
            state_hash = response_from_rc[3]
        else:
            raise DatabaseException(f"get_calc_response_from_rc invalid version: {version}")

        return iscore, block_height, state_hash

    def get_tx_index(self, start_calc: bool) -> int:
        tx_index: int = -1
        if start_calc:
            return tx_index
        else:
            return self._db_iiss_tx_index

    @staticmethod
    def put(batch: list, iiss_data: 'Data'):
        Logger.debug(tag=IISS_LOG_TAG, msg=f"put data: {str(iiss_data)}")
        batch.append(iiss_data)

    def commit(self, iiss_wal: 'IissWAL'):
        self._db.write_batch(iiss_wal)
        self._db_iiss_tx_index = iiss_wal.final_tx_index
        Logger.info(tag=IISS_LOG_TAG, msg=f"final_tx_index={iiss_wal.final_tx_index}")

    # todo: naming
    def _put_version_and_revision(self, revision: int):
        version: int = get_rc_version(revision)
        version_and_revision: bytes = MsgPackForDB.dumps([version, revision])
        self._db.put(self.KEY_FOR_VERSION_AND_REVISION, version_and_revision)

    def get_version_and_revision(self) -> Tuple[int, int]:
        return get_version_and_revision(self._db)

    def _load_last_transaction_index(self) -> int:
        encoded_last_index: Optional[bytes] = self._db.get(self.KEY_FOR_GETTING_LAST_TRANSACTION_INDEX)
        if encoded_last_index is None:
            return -1
        else:
            return int.from_bytes(encoded_last_index, DATA_BYTE_ORDER)

    @staticmethod
    def _rename_db(old_db_path: str, new_db_path: str):
        if os.path.exists(old_db_path) and not os.path.exists(new_db_path):
            shutil.move(old_db_path, new_db_path)
            Logger.info(tag=IISS_LOG_TAG, msg=f"Rename db: {old_db_path} -> {new_db_path}")
        else:
            raise DatabaseException("Cannot create IISS DB because of invalid path. Check both IISS "
                                    "current DB path and IISS DB path")

    def replace_db(self, block_height: int) -> 'RewardCalcDBInfo':
        """
        1. Rename current_db to standby_db_{block_height}
        2. Create a new current_db for the next calculation period

        :param block_height: End block height of the current calc period
        :return:
        """

        # rename current db -> standby db
        assert block_height > 0

        self._db.close()

        standby_db_path: str = self.rename_current_db_to_standby_db(self._path, block_height)
        self._db = self.create_current_db(self._path)

        return RewardCalcDBInfo(standby_db_path, block_height)

    @classmethod
    def finalize_iiss_db(cls,
                         prev_end_bh: int,
                         current_db: 'KeyValueDatabase',
                         prev_db_path: str):
        """
        Finalize iiss db before sending to reward calculator (i.e. RC). Process is below
            1. Move last Block produce data to previous iiss_db which is to be sent to RC
            2. db compaction

        :param prev_end_bh: end block height of previous term
        :param current_db: newly created db
        :param prev_db_path: iiss_db path which is to be finalized and sent to RC (must has been closed)
        :return:
        """
        bp_key: bytes = make_block_produce_info_key(prev_end_bh)
        prev_db: 'KeyValueDatabase' = KeyValueDatabase.from_path(prev_db_path)
        cls._move_data_from_current_db_to_prev_db(bp_key,
                                                  current_db,
                                                  prev_db)
        prev_db.close()
        cls._process_db_compaction(prev_db_path)

    @classmethod
    def _move_data_from_current_db_to_prev_db(cls,
                                              key: bytes,
                                              current_db: 'KeyValueDatabase',
                                              prev_db: 'KeyValueDatabase'):
        value: Optional[bytes] = current_db.get(key)
        if value is None:
            return

        current_db.delete(key)
        prev_db.put(key, value)

    @classmethod
    def _process_db_compaction(cls, path: str):
        """
        There is compatibility issue between C++ levelDB and go levelDB.
        To solve it, should make DB being compacted before reading (from RC).
        :param path: DB path to compact
        :return:
        """
        db = KeyValueDatabase.from_path(path)
        db.close()

    @classmethod
    def create_current_db(cls, rc_data_path: str) -> 'KeyValueDatabase':
        current_db_path = os.path.join(rc_data_path, cls.CURRENT_IISS_DB_NAME)

        Logger.info(tag=IISS_LOG_TAG, msg=f"Create new current_db")
        return KeyValueDatabase.from_path(current_db_path, create_if_missing=True)

    @classmethod
    def rename_current_db_to_standby_db(cls, rc_data_path: str, block_height: int) -> str:
        current_db_path: str = os.path.join(rc_data_path, cls.CURRENT_IISS_DB_NAME)
        standby_db_name: str = cls.get_standby_rc_db_name(block_height)
        standby_db_path: str = os.path.join(rc_data_path, standby_db_name)

        cls._rename_db(current_db_path, standby_db_path)

        return standby_db_path

    @classmethod
    def rename_standby_db_to_iiss_db(cls, standby_db_path: str) -> str:
        # After change the db name, reward calc manage this db (icon service does not have a authority)
        iiss_db_path: str = cls.IISS_RC_DB_NAME_PREFIX.\
            join(standby_db_path.rsplit(cls.STANDBY_IISS_DB_NAME_PREFIX, 1))

        cls._rename_db(standby_db_path, iiss_db_path)

        return iiss_db_path

    def get_total_elected_prep_delegated_snapshot(self) -> int:
        """
        total_elected_prep_delegated_snapshot =
            the delegated amount which the elected P-Reps received at the beginning of this term
            - the delegated amount which unregistered P-Reps received in this term

        This function is only intended for state backward compatibility
        and not used any more after revision is set to 7.
        """

        unreg_preps: Set['Address'] = set()
        db = self._db.get_sub_db(TxData.PREFIX)
        for k, v in db.iterator():
            data: 'TxData' = TxData.from_bytes(v)
            if data.type == TxType.PREP_UNREGISTER:
                unreg_preps.add(data.address)

        db = self._db.get_sub_db(PRepsData.PREFIX)
        preps: Optional[List['DelegationInfo']] = None
        for k, v in db.iterator():
            data: 'PRepsData' = PRepsData.from_bytes(k, v)
            preps = data.prep_list
            break

        ret = 0
        if preps:
            for info in preps:
                if info.address not in unreg_preps:
                    ret += info.value

        Logger.info(tag=IISS_LOG_TAG,
                    msg=f"get_total_elected_prep_delegated_snapshot load: {ret}")

        return ret


class IissDBNameRefactor(object):
    """Change iiss_db name: remove revision from iiss_db name

    """
    _DB_NAME_PREFIX = Storage.IISS_RC_DB_NAME_PREFIX

    @classmethod
    def run(cls, rc_data_path: str) -> int:
        ret = 0

        with os.scandir(rc_data_path) as it:
            for entry in it:
                if entry.is_dir() and entry.name.startswith(cls._DB_NAME_PREFIX):
                    new_name: str = cls._get_db_name_without_revision(entry.name)
                    if not new_name:
                        Logger.info(
                            tag=IISS_LOG_TAG,
                            msg=f"Refactoring iiss_db name has been already done: old={entry.name} "
                                f"rc_data_path={rc_data_path}")
                        break

                    cls._change_db_name(rc_data_path, entry.name, new_name)
                    ret += 1

        return ret

    @classmethod
    def _change_db_name(cls, rc_data_path: str, old_name: str, new_name: str):
        if old_name == new_name:
            return

        src_path: str = os.path.join(rc_data_path, old_name)
        dst_path: str = os.path.join(rc_data_path, new_name)

        try:
            shutil.move(src_path, dst_path)
            Logger.info(tag=IISS_LOG_TAG, msg=f"Renaming iiss_db_name succeeded: old={old_name} new={new_name}")
        except BaseException as e:
            Logger.error(tag=IISS_LOG_TAG,
                         msg=f"Failed to rename iiss_db_name: old={old_name} new={new_name} "
                             f"path={rc_data_path} exception={str(e)}")

    @classmethod
    def _get_db_name_without_revision(cls, name: str) -> Optional[str]:
        # items[0]: block_height, items[1]: revision
        items: List[str] = name[len(cls._DB_NAME_PREFIX) + 1:].split("_")
        if len(items) == 1:
            # No need to rename
            return None

        return f"{cls._DB_NAME_PREFIX}_{items[0]}"



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import unique, IntEnum
from struct import Struct
from typing import Optional

from .exception import InvalidParamsException
from ..icon_constant import DATA_BYTE_ORDER, DEFAULT_BYTE_SIZE, Revision
from ..utils import bytes_to_hex
from ..utils.msgpack_for_db import MsgPackForDB


@unique
class BlockVersion(IntEnum):
    STRUCT = 0
    # change to msg pack and add 'step_used' field
    MSG_PACK = 1


class Block(object):
    """Block Information included in IconScoreContext
    """
    _VERSION = BlockVersion.MSG_PACK
    _STRUCT_PACKED_BYTES_SIZE = 129
    # leveldb account value structure (bigendian, 1 + 32 + 32 + 32 + 32 bytes)
    # version(1)
    # | height(DEFAULT_BYTE_SIZE)
    # | hash(DEFAULT_BYTE_SIZE)
    # | timestamp(DEFAULT_BYTE_SIZE)
    # | prev_hash(DEFAULT_BYTE_SIZE)

    _struct = Struct(f'>B{DEFAULT_BYTE_SIZE}s{DEFAULT_BYTE_SIZE}s{DEFAULT_BYTE_SIZE}s{DEFAULT_BYTE_SIZE}s')

    def __init__(self,
                 block_height: int,
                 block_hash: Optional[bytes],
                 timestamp: int,
                 prev_hash: Optional[bytes],
                 cumulative_fee: int = 0) -> None:
        """Constructor

        :param block_height: block height
        :param block_hash: block hash
        :param timestamp: block timestamp
        :param prev_hash: prev block hash
        :param cumulative_fee: cumulative_fee
        """
        self._height = block_height
        self._hash = block_hash
        # unit: microsecond
        self._timestamp = timestamp
        self._prev_hash = prev_hash
        # set default value for compatibility with t-bears
        self.cumulative_fee = cumulative_fee

    def to_dict(self, casing: Optional[callable] = None) -> dict:
        """
        Returns properties as `dict`
        :return: a dict
        """
        new_dict = {}
        for key, value in self.__dict__.items():
            if key.startswith("_"):
                key = key[1:]
            new_dict[casing(key) if casing else key] = value

        return new_dict

    def __str__(self) -> str:
        return f"Block(height={self._height}, " \
               f"hash={bytes_to_hex(self._hash)}, " \
               f"prev_hash={bytes_to_hex(self._prev_hash)}, " \
               f"timestamp={self._timestamp}, " \
               f"cumulative_fee={self.cumulative_fee})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, Block) \
               and self._height == other._height \
               and self._timestamp == other._timestamp \
               and self._prev_hash == other._prev_hash \
               and self.cumulative_fee == other.cumulative_fee

    @property
    def height(self) -> int:
        return self._height

    @property
    def hash(self) -> bytes:
        return self._hash

    @property
    def timestamp(self) -> int:
        return self._timestamp

    @property
    def prev_hash(self) -> bytes:
        return self._prev_hash

    @staticmethod
    def from_dict(params: dict):
        block_height = params.get('blockHeight')
        block_hash = params.get('blockHash')
        timestamp = params.get('timestamp', 0)
        prev_hash = params.get('prevBlockHash', b'\x00' * 32)

        return Block(block_height=block_height,
                     block_hash=block_hash,
                     timestamp=timestamp,
                     prev_hash=prev_hash,
                     cumulative_fee=0)

    @staticmethod
    def from_block(block: 'Block'):
        block_height = block.height
        block_hash = block.hash
        timestamp = block.timestamp
        prev_hash = block.prev_hash
        cumulative_fee = block.cumulative_fee
        return Block(block_height, block_hash, timestamp, prev_hash, cumulative_fee)

    @staticmethod
    def from_bytes(buf: bytes) -> 'Block':
        if len(buf) == Block._STRUCT_PACKED_BYTES_SIZE and buf[0] == BlockVersion.STRUCT:
            return Block._from_struct_packed_bytes(buf)
        else:
            return Block._from_msg_packed_bytes(buf)

    @staticmethod
    def _from_struct_packed_bytes(buf: bytes) -> 'Block':
        """Create Account object from bytes data

        :param buf: (bytes) bytes data including Account information
        :return: (Account) account object
        """
        byteorder = DATA_BYTE_ORDER

        version, block_height_bytes, block_hash_bytes, \
            timestamp_bytes, block_prev_hash_bytes = Block._struct.unpack(buf)

        block_height = int.from_bytes(block_height_bytes, byteorder)
        block_hash = block_hash_bytes
        timestamp = int.from_bytes(timestamp_bytes, byteorder)
        byte_prev_hash = block_prev_hash_bytes

        if int(bytes.hex(byte_prev_hash), 16) == 0:
            byte_prev_hash = None
        prev_block_hash = byte_prev_hash

        block = Block(block_height, block_hash, timestamp, prev_block_hash, 0)
        return block

    @staticmethod
    def _from_msg_packed_bytes(buf: bytes) -> 'Block':
        data: list = MsgPackForDB.loads(buf)
        version: int = data[0]

        assert version <= Block._VERSION

        if version != BlockVersion.MSG_PACK:
            raise InvalidParamsException(f"Invalid block version: {version}")

        return Block(block_height=data[1],
                     block_hash=data[2],
                     timestamp=data[3],
                     prev_hash=data[4],
                     cumulative_fee=data[5])

    def to_bytes(self, revision: int) -> bytes:
        if revision >= Revision.IISS.value:
            return self._to_msg_packed_bytes()
        else:
            return self._to_struct_packed_bytes()

    def _to_msg_packed_bytes(self) -> bytes:
        data = [
            BlockVersion.MSG_PACK.value,
            self._height,
            self._hash,
            self._timestamp,
            self._prev_hash,
            self.cumulative_fee
        ]
        return MsgPackForDB.dumps(data)

    def _to_struct_packed_bytes(self) -> bytes:
        """Convert block object to bytes

        :return: data including information of block object
        """
        byteorder = DATA_BYTE_ORDER
        # for extendability
        block_height_bytes = self._height.to_bytes(DEFAULT_BYTE_SIZE, byteorder)
        block_hash_bytes = self._hash
        timestamp_bytes = self._timestamp.to_bytes(DEFAULT_BYTE_SIZE, byteorder)

        tmp_prev_hash = self._prev_hash
        if tmp_prev_hash is None:
            tmp_prev_hash = bytes(DEFAULT_BYTE_SIZE)
        prev_block_hash_bytes = tmp_prev_hash

        return Block._struct.pack(
            BlockVersion.STRUCT,
            block_height_bytes,
            block_hash_bytes,
            timestamp_bytes,
            prev_block_hash_bytes)


# This predefined block is used to fix context.block.height access error before genesis block is synchronized.
NULL_BLOCK = Block(block_height=-1, block_hash=None, timestamp=0, prev_hash=None)



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from asyncio import StreamReader, StreamWriter
from typing import Optional

from iconcommons import Logger
from .message import MessageType, Request
from .message_queue import MessageQueue
from .message_unpacker import MessageUnpacker

_TAG = "RCP"


class IPCServer(object):
    def __init__(self):
        self._running = False
        self._loop = None
        self._path = None
        self._queue: Optional['MessageQueue'] = None
        self._unpacker: Optional['MessageUnpacker'] = MessageUnpacker()
        self._tasks = []

    def open(self, loop,  message_queue: 'MessageQueue', path: str):
        Logger.info(tag=_TAG, msg="open() start")

        assert loop
        assert message_queue
        assert isinstance(path, str)

        self._loop = loop
        self._queue = message_queue
        self._path = path

        Logger.info(tag=_TAG, msg="open() end")

    def start(self):
        Logger.info(tag=_TAG, msg="start() start")

        if self._running:
            return

        self._running = True
        co = asyncio.start_unix_server(self._on_accepted, self._path)
        self._loop.run_until_complete(co)

        Logger.info(tag=_TAG, msg="start() end")

    def stop(self):
        Logger.info(tag=_TAG, msg="stop() start")

        if not self._running:
            return

        self._running = False

        for t in self._tasks:
            t.cancel()

        Logger.info(tag=_TAG, msg="stop() end")

    def close(self):
        Logger.info(tag=_TAG, msg="close() start")

        self._loop = None
        self._unpacker = None

        Logger.info(tag=_TAG, msg="close() end")

    def _on_accepted(self, reader: 'StreamReader', writer: 'StreamWriter'):
        Logger.info(tag=_TAG, msg=f"on_accepted() start: {reader} {writer}")

        self._tasks.append(asyncio.ensure_future(self._on_send(writer)))
        self._tasks.append(asyncio.ensure_future(self._on_recv(reader)))

        Logger.info(tag=_TAG, msg="on_accepted() end")

    async def _on_send(self, writer: 'StreamWriter'):
        Logger.info(tag=_TAG, msg="_on_send() start")

        while self._running:
            try:
                request: 'Request' = await self._queue.get()
                self._queue.task_done()

                if request.msg_type == MessageType.NONE:
                    # Stopping IPCServer
                    break

                data: bytes = request.to_bytes()
                Logger.debug(tag=_TAG, msg=f"on_send(): data({data.hex()}")
                Logger.info(tag=_TAG, msg=f"Sending Data : {request}")
                writer.write(data)
                await writer.drain()

            except asyncio.CancelledError:
                # task got cancel request. stop service
                break
            except BaseException as e:
                Logger.warning(tag=_TAG, msg=str(e))

        writer.close()

        Logger.info(tag=_TAG, msg="_on_send() end")

    async def _on_recv(self, reader: 'StreamReader'):
        Logger.info(tag=_TAG, msg="_on_recv() start")

        while self._running:
            try:
                data: bytes = await reader.read(1024)
                if not isinstance(data, bytes) or len(data) == 0:
                    break

                Logger.debug(tag=_TAG, msg=f"_on_recv(): data({data.hex()})")

                self._unpacker.feed(data)

                for response in self._unpacker:
                    Logger.info(tag=_TAG, msg=f"Received Data : {response}")
                    self._queue.message_handler(response)

            except asyncio.CancelledError:
                # task got cancel request. stop service
                break
            except BaseException as e:
                Logger.warning(tag=_TAG, msg=str(e))

        Logger.info(tag=_TAG, msg="_on_recv() end")



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from ..base.ComponentBase import StorageBase
from ..base.exception import FatalException
from ..icon_constant import IISS_MAX_REWARD_RATE, ConfigKey
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from ..database.db import ContextDatabase


class Storage(StorageBase):
    PREFIX: bytes = b'iiss'

    IISS_META_DATA_KEY: bytes = PREFIX + b'md'
    REWARD_RATE_KEY: bytes = PREFIX + b'rr'
    TOTAL_STAKE_KEY: bytes = PREFIX + b'ts'

    CALC_END_BLOCK_HEIGHT_KEY: bytes = PREFIX + b'cebh'
    CALC_PERIOD_KEY: bytes = PREFIX + b'pk'

    def __init__(self, db: 'ContextDatabase'):
        super().__init__(db)
        self._meta_data: Optional['IISSMetaData'] = None

    def open(self,
             context: 'IconScoreContext',
             iiss_meta_data: dict,
             calc_period: int):
        self.check_config_before_init(iiss_meta_data)

        meta_data: Optional['IISSMetaData'] = self.get_meta_data(context)
        if meta_data is None:
            meta_data = IISSMetaData(reward_min=iiss_meta_data[ConfigKey.REWARD_MIN],
                                     reward_max=iiss_meta_data[ConfigKey.REWARD_MAX],
                                     reward_point=iiss_meta_data[ConfigKey.REWARD_POINT],
                                     lock_min=iiss_meta_data[ConfigKey.UN_STAKE_LOCK_MIN],
                                     lock_max=iiss_meta_data[ConfigKey.UN_STAKE_LOCK_MAX])
            self.put_meta_data(context, meta_data)

        if self.get_calc_period(context) is None:
            self.put_calc_period(context, calc_period)

        self._meta_data: 'IISSMetaData' = meta_data

    def rollback(self, context: 'IconScoreContext', block_height: int, block_hash: bytes):
        pass

    @property
    def reward_min(self) -> int:
        return self._meta_data.reward_min

    @property
    def reward_max(self) -> int:
        return self._meta_data.reward_max

    @property
    def reward_point(self) -> int:
        return self._meta_data.reward_point

    @property
    def lock_min(self) -> int:
        return self._meta_data.lock_min

    @property
    def lock_max(self) -> int:
        return self._meta_data.lock_max

    @staticmethod
    def check_config_before_init(iiss_variable: dict):
        for k, v in iiss_variable.items():
            if k.startswith('reward') and not 0 < v <= IISS_MAX_REWARD_RATE:
                raise FatalException(f"Out of reward rate range: 0 < {v} <= {IISS_MAX_REWARD_RATE}")

    def put_meta_data(self, context: 'IconScoreContext', constant: 'IISSMetaData'):
        self._db.put(context, self.IISS_META_DATA_KEY, constant.to_bytes())

    def get_meta_data(self, context: 'IconScoreContext') -> Optional['IISSMetaData']:
        value: Optional[bytes] = self._db.get(context, self.IISS_META_DATA_KEY)
        if value:
            return IISSMetaData.from_bytes(value)
        else:
            return None

    def put_reward_rate(self, context: 'IconScoreContext', reward_rate: 'RewardRate'):
        self._db.put(context, self.REWARD_RATE_KEY, reward_rate.to_bytes())

    def get_reward_rate(self, context: 'IconScoreContext') -> Optional['RewardRate']:
        value: Optional[bytes] = self._db.get(context, self.REWARD_RATE_KEY)
        if value:
            return RewardRate.from_bytes(value)
        else:
            return RewardRate(reward_prep=None)

    def put_total_stake(self, context: 'IconScoreContext', total_stake: int):
        version = 0
        data: bytes = MsgPackForDB.dumps([version, total_stake])
        self._db.put(context, self.TOTAL_STAKE_KEY, data)

    def get_total_stake(self, context: 'IconScoreContext') -> int:
        value: bytes = self._db.get(context, self.TOTAL_STAKE_KEY)
        if value:
            data = MsgPackForDB.loads(value)
            version: int = data[0]
            assert version == 0
            total_stake: int = data[1]
            return total_stake
        else:
            return 0

    def put_end_block_height_of_calc(self, context: 'IconScoreContext', calc_block_height: int):
        version = 0
        data: list = [version, calc_block_height]
        value: bytes = MsgPackForDB.dumps(data)
        self._db.put(context, self.CALC_END_BLOCK_HEIGHT_KEY, value)

    def get_end_block_height_of_calc(self, context: 'IconScoreContext') -> Optional[int]:
        value: bytes = self._db.get(context, self.CALC_END_BLOCK_HEIGHT_KEY)
        if value:
            data: list = MsgPackForDB.loads(value)
            version: int = data[0]
            assert version == 0
            calc_block_height: int = data[1]
            return calc_block_height
        return None

    def put_calc_period(self, context: 'IconScoreContext', calc_period: int):
        version = 0
        data: list = [version, calc_period]
        value: bytes = MsgPackForDB.dumps(data)
        self._db.put(context, self.CALC_PERIOD_KEY, value)

    def get_calc_period(self, context: 'IconScoreContext') -> Optional[int]:
        value: bytes = self._db.get(context, self.CALC_PERIOD_KEY)
        if value:
            data: list = MsgPackForDB.loads(value)
            version: int = data[0]
            assert version == 0
            calc_period: int = data[1]
            return calc_period
        return None


class RewardRate:
    _VERSION = 0

    def __init__(self,
                 reward_prep: Optional[int]):
        # todo: if eep, dapp is added, eep_rate, dapp rate is added as a member variable
        self.reward_prep: Optional[int] = reward_prep

    def __str__(self):
        return f"reward_prep={self.reward_prep}"

    @classmethod
    def from_bytes(cls, buf: bytes) -> 'RewardRate':
        data: list = MsgPackForDB.loads(buf)
        version = data[0]
        assert version == cls._VERSION

        return cls(*data[1:])

    def to_bytes(self):
        data: list = [
            self._VERSION,
            self.reward_prep
        ]
        return MsgPackForDB.dumps(data)


class IISSMetaData:
    _VERSION = 0

    def __init__(self,
                 reward_min: int,
                 reward_max: int,
                 reward_point: int,
                 lock_min: int,
                 lock_max: int):
        self.reward_min: int = reward_min
        self.reward_max: int = reward_max
        self.reward_point: int = reward_point
        self.lock_min: int = lock_min
        self.lock_max: int = lock_max

    def __str__(self):
        return f"reward_min={self.reward_min}, " \
               f"reward_max={self.reward_max}, " \
               f"reward_point={self.reward_point}, " \
               f"lock_min={self.lock_min}, " \
               f"lock_max={self.lock_max}"

    @classmethod
    def from_bytes(cls, buf: bytes) -> 'IISSMetaData':
        data: list = MsgPackForDB.loads(buf)
        version = data[0]
        assert version == cls._VERSION

        return cls(*data[1:])

    def to_bytes(self):
        data: list = [
            self._VERSION,
            self.reward_min,
            self.reward_max,
            self.reward_point,
            self.lock_min,
            self.lock_max
        ]
        return MsgPackForDB.dumps(data)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List, Optional

from .icon_score_event_log import EventLog
from ..base.address import Address
from ..base.block import Block
from ..base.exception import ExceptionCode
from ..icon_constant import DATA_BYTE_ORDER
from ..utils import bytes_to_hex
from ..utils.bloom import BloomFilter

if TYPE_CHECKING:
    from ..base.transaction import Transaction


class TransactionResult(object):
    """ A DataClass of a transaction result.
    """

    SUCCESS = 1
    FAILURE = 0

    class Failure(object):
        def __init__(self, code: int, message: str):
            """MUST check arguments type strictly

            :param code: error code
            :param message: error message
            """
            if type(code) != int:
                code = ExceptionCode.SYSTEM_ERROR.value
            if type(message) != str:
                message = 'Invalid argument: message is not a string'

            self.code = code
            self.message = message

    def __init__(
            self,
            tx: 'Transaction',
            block: 'Block',
            to: Optional['Address'] = None,
            score_address: Optional['Address'] = None,
            step_used: int = 0,
            step_price: int = 0,
            cumulative_step_used: int = 0,
            event_logs: Optional[List['EventLog']] = None,
            logs_bloom: Optional[BloomFilter] = None,
            status: int = FAILURE) -> None:
        """Constructor

        :param tx: transaction
        :param block: a block that the transaction belongs to
        :param to: a recipient address
        :param score_address:hex string that represent the contract address
            if the transaction`s target is contract
        :param step_used: the amount of steps used in the transaction
        :param event_logs: the amount of steps used in the transaction
        :param logs_bloom: bloom filter data of event logs
        :param status: a status of result. 1 (success) or 0 (failure)
        """
        self.tx_hash = tx.hash
        self.block_height = block.height
        self.block_hash = block.hash
        self.tx_index = tx.index
        self.to = to
        self.score_address = score_address
        self.step_used = step_used
        self.step_price = step_price
        self.cumulative_step_used = cumulative_step_used
        self.event_logs = event_logs
        self.logs_bloom = logs_bloom
        self.status = status

        # Details of the used step. This is set if the SCORE pays fees.
        # Otherwise left as `None` and not passed to transaction result.
        self.step_used_details: Optional[dict] = None

        # failure object which has code(int) and message(str) attributes
        # It is only available on self.status == FAILURE
        self.failure = None

        # Traces are managed in TransactionResult but not passed to chain engine
        self.traces = None

    def __str__(self) -> str:
        def func():
            for k, v in self.__dict__.items():
                if isinstance(v, bytes):
                    v = bytes_to_hex(v)
                yield f'{k}: {v}'

        return '\n'.join(func())

    def to_dict(self, casing: Optional[callable] = None) -> dict:
        """
        Returns properties as `dict`
        :return: a dict
        """
        new_dict = {}
        for key, value in self.__dict__.items():
            # Excludes properties which have `None` value
            if value is None:
                continue

            new_key = casing(key) if casing else key
            if key == 'event_logs':
                new_dict[new_key] = [v.to_dict(casing) for v in value if
                                     isinstance(v, EventLog)]
            elif isinstance(value, BloomFilter):
                new_dict[new_key] = int(value).to_bytes(256, byteorder=DATA_BYTE_ORDER)
            elif key == 'failure':
                if self.status == self.FAILURE:
                    new_dict[new_key] = {
                        'code': value.code,
                        'message': value.message
                    }
            elif key == 'step_used_details':
                assert isinstance(value, dict)
                step_used_details = {}
                new_dict[new_key] = step_used_details

                for address in value:
                    step_used_details[str(address)] = value[address]
            elif key == 'traces':
                # traces are excluded from dict property
                continue
            else:
                new_dict[new_key] = value

        return new_dict



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from enum import IntEnum
from typing import Any

from msgpack import dumps as msgpack_dumps, loads as msgpack_loads, ExtType as msgpack_extType

from . import int_to_bytes, bytes_to_int
from ..base.address import Address


# you should override if you want to parse custom type
# class CustomCodec(Codec):
#     class CustomType(IntEnum):
#         # NONE = 0
#         # BIG_INT = 1
#         # ADDRESS = 2
#
#         CUSTOM = 3
#
#     @classmethod
#     def encode(cls, obj: Any) -> Any:
#         return obj
#
#     @classmethod
#     def decode(cls, t: int, b: bytes) -> Any:
#         return msgpack_extType(t, b)


class Codec(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def encode(cls, obj: Any) -> Any:
        pass

    @classmethod
    @abstractmethod
    def decode(cls, t: int, b: bytes) -> Any:
        pass


class MsgPackForDB(object):
    class BaseCodec(Codec):
        @classmethod
        def encode(cls, obj: Any) -> Any:
            return obj

        @classmethod
        def decode(cls, t: int, b: bytes) -> Any:
            return msgpack_extType(t, b)

    # you should assign CustomCodec if you want to parse custom type
    _codec: 'Codec' = BaseCodec()

    class BaseType(IntEnum):
        NONE = 0
        BIG_INT = 1
        ADDRESS = 2

    @classmethod
    def _encode(cls, obj: Any) -> Any:
        if isinstance(obj, int):
            return msgpack_extType(cls.BaseType.BIG_INT, int_to_bytes(obj))
        elif isinstance(obj, Address):
            return msgpack_extType(cls.BaseType.ADDRESS, obj.to_bytes_including_prefix())
        else:
            return cls._codec.encode(obj)

    @classmethod
    def _decode(cls, t: int, b: bytes) -> Any:
        if t == cls.BaseType.BIG_INT:
            return bytes_to_int(b)
        elif t == cls.BaseType.ADDRESS:
            return Address.from_bytes_including_prefix(b)
        else:
            return cls._codec.decode(t, b)

    @classmethod
    def dumps(cls, data: Any) -> bytes:
        return msgpack_dumps(data, default=cls._encode, use_bin_type=True, strict_types=True)

    @classmethod
    def loads(cls, data: bytes) -> list:
        return msgpack_loads(data, ext_hook=cls._decode, raw=False, strict_map_key=False)



# -*- coding: utf-8 -*-

from __future__ import annotations

__all__ = "UnstakePatcher"

import importlib.resources
import json
from enum import IntEnum
from typing import TYPE_CHECKING, List, Dict, Any

from iconcommons.logger import Logger

from .storage import AccountPartFlag
from ..base.address import Address, SYSTEM_SCORE_ADDRESS
from ..iconscore.icon_score_event_log import EventLogEmitter
from ..icx.coin_part import CoinPartFlag

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from ..icx.coin_part import CoinPart
    from ..icx.stake_part import StakePart

TAG = "UNSTAKE"
INVALID_EXPIRED_UNSTAKES_FILENAME = "invalid_expired_unstakes.json"


class Unstake(object):
    def __init__(self, amount: int, block_height: int):
        self._amount = amount
        self._block_height = block_height

    def __str__(self) -> str:
        return f"{self.to_list()}"

    @property
    def amount(self) -> int:
        return self._amount

    @property
    def block_height(self) -> int:
        return self._block_height

    def to_list(self) -> List[int]:
        return [self._amount, self._block_height]

    @classmethod
    def from_list(cls, data: List[int]) -> Unstake:
        return cls(amount=data[0], block_height=data[1])


class Target(object):
    def __init__(self, address: Address, unstakes: List[Unstake]):
        self._address = address
        self._unstakes: List[Unstake] = unstakes

    def __len__(self) -> int:
        return len(self._unstakes)

    def __str__(self):
        return (
            f"address={self._address} "
            f"total_unstake={self.total_unstake} "
            f"unstakes={self._unstakes}"
        )

    @property
    def address(self) -> Address:
        return self._address

    @property
    def total_unstake(self) -> int:
        return sum(unstake.amount for unstake in self._unstakes)

    @property
    def unstakes(self) -> List[Unstake]:
        return self._unstakes

    def add_unstake(self, amount: int, block_height: int):
        self._unstakes.append(Unstake(amount, block_height))

    def to_dict(self) -> Dict[str, Any]:
        return {
            "address": str(self._address),
            "total_unstake": self.total_unstake,
            "unstakes": [unstake.to_list() for unstake in self._unstakes],
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Target:
        unstakes: List[Unstake] = [Unstake.from_list(i) for i in data["unstakes"]]

        return cls(address=Address.from_string(data["address"]), unstakes=unstakes)


class Result(IntEnum):
    FALSE = 0
    REMOVABLE_V0 = 1
    REMOVABLE_V1 = 2


class UnstakePatcher(object):

    def __init__(self, targets: List[Target]):
        self._targets = targets

        self._success_targets: List[Target] = []
        self._failure_targets: List[Target] = []
        self._success_unstake = 0
        self._failure_unstake = 0

    @classmethod
    def _load(cls, path: str) -> Dict[str, Any]:
        if isinstance(path, str):
            with open(path, "rt") as f:
                json_text = f.read()
        else:
            json_text = importlib.resources.read_text(
                "iconservice.res", INVALID_EXPIRED_UNSTAKES_FILENAME
            )

        return json.loads(json_text)

    def run(self, context: 'IconScoreContext'):
        Logger.info(tag=TAG, msg="UnstakePatcher.run() start")

        storage = context.storage.icx

        for target in self._targets:
            try:
                address = target.address
                coin_part = storage.get_part(context, AccountPartFlag.COIN, address)
                stake_part = storage.get_part(context, AccountPartFlag.STAKE, address)

                result: Result = self._check_removable(coin_part, stake_part, target)
                if result == Result.FALSE:
                    self._add_failure_item(target)
                else:
                    if result == Result.REMOVABLE_V0:
                        stake_part = self._remove_invalid_expired_unstakes_v0(stake_part, target)
                    else:
                        stake_part = self._remove_invalid_expired_unstakes_v1(stake_part, target)

                    assert stake_part.is_dirty()
                    storage.put_stake_part(context, address, stake_part)
                    self._emit_event_log(context, target)
                    self._add_success_item(target)
            except BaseException as e:
                # Although some unexpected errors happen, keep going
                Logger.exception(tag=TAG, msg=str(e))

        Logger.info(tag=TAG, msg="UnstakePatcher.run() end")

    @classmethod
    def _check_removable(
        cls, coin_part: 'CoinPart', stake_part: 'StakePart', target: Target
    ) -> Result:

        if CoinPartFlag.HAS_UNSTAKE not in coin_part.flags:
            if stake_part.unstake_block_height > 0:
                if cls._is_removable_v0(stake_part, target):
                    return Result.REMOVABLE_V0
            else:
                if cls._is_removable_v1(stake_part, target):
                    return Result.REMOVABLE_V1

        return Result.FALSE

    @classmethod
    def _is_removable_v0(cls, stake_part: 'StakePart', target: Target) -> bool:
        """Inspect stake_part.unstake and stake_part.unstake_block_height

        :param target:
        :param stake_part:
        :return:
        """
        if len(target.unstakes) != 1:
            return False
        if len(stake_part.unstakes_info) != 0:
            Logger.error(
                tag=TAG,
                msg=f"Invalid stake_part.unstakes_info: {stake_part.unstakes_info}"
            )
            return False

        unstake: 'Unstake' = target.unstakes[0]
        return (
            unstake.amount == stake_part.unstake
            and unstake.block_height == stake_part.unstake_block_height
        )

    @classmethod
    def _is_removable_v1(cls, stake_part: 'StakePart', target: Target) -> bool:
        """Inspect stake_part.unstakes_info

        :param stake_part:
        :param target:
        :return:
        """
        unstakes_info: List[List[int, int]] = stake_part.unstakes_info
        if (
                len(unstakes_info) == 0
                or stake_part.unstake > 0
                or stake_part.unstake_block_height > 0
        ):
            return False

        for info, unstake in zip(stake_part.unstakes_info, target.unstakes):
            if not (info[0] == unstake.amount and info[1] == unstake.block_height):
                return False

        return True

    @classmethod
    def _remove_invalid_expired_unstakes_v0(cls, stake_part: 'StakePart', target: Target) -> 'StakePart':
        """Remove invalid expired unstakes from stake_part.unstake and stake_part.unstake_block_height

        :param stake_part:
        :param target:
        :return:
        """
        assert len(stake_part.unstakes_info) == 0
        assert len(target.unstakes) == 1

        unstake: Unstake = target.unstakes[0]
        assert stake_part.unstake == unstake.amount
        assert stake_part.unstake_block_height == unstake.block_height

        stake_part.cleanup_old_format_unstake()
        assert stake_part.unstake == 0
        assert stake_part.unstake_block_height == 0

        Logger.info(tag=TAG, msg=f"remove_invalid_expired_unstakes_v0: {target}")

        return stake_part

    @classmethod
    def _remove_invalid_expired_unstakes_v1(cls, stake_part: 'StakePart', target: Target) -> 'StakePart':
        """Remove invalid expired unstakes from stake_part.unstakes_info

        :param stake_part:
        :param target:
        :return:
        """
        assert stake_part.unstake == 0
        assert stake_part.unstake_block_height == 0

        for unstake in target.unstakes:
            amount, unstake_block_height = stake_part.unstakes_info.pop(0)

            assert amount == unstake.amount
            assert unstake_block_height == unstake.block_height

        Logger.info(tag=TAG, msg=f"remove_invalid_expired_unstakes_v1: {target}")

        stake_part.set_dirty(True)
        return stake_part

    def _add_success_item(self, target: Target):
        self._success_targets.append(target)
        self._success_unstake += target.total_unstake

    def _add_failure_item(self, target: Target):
        self._failure_targets.append(target)
        self._failure_unstake += target.total_unstake

    @classmethod
    def _emit_event_log(cls, context: 'IconScoreContext', target: Target):
        for unstake in target.unstakes:
            EventLogEmitter.emit_event_log(
                context=context,
                event_signature="InvalidUnstakeFixed(Address,int,int)",
                score_address=SYSTEM_SCORE_ADDRESS,
                arguments=[target.address, unstake.amount, unstake.block_height],
                indexed_args_count=1,
                fee_charge=False
            )

    def write_result(self, path: str):
        Logger.info(tag=TAG, msg=f"UnstakePatcher.write_result() start: {path}")

        total_unstake = self._success_unstake + self._failure_unstake

        Logger.info(
            tag=TAG,
            msg="Invalid expired unstakes management result: "
            f"total_unstake={total_unstake} "
            f"success_unstake={self._success_unstake} "
            f"failure_unstake={self._failure_unstake} "
            f"total_items={len(self._targets)} "
            f"success_items={len(self._success_targets)} "
            f"failure_items={len(self._failure_targets)}",
        )

        try:
            report = {
                # Unstake amount
                "total_unstake": total_unstake,
                "success_unstake": self._success_unstake,
                "failure_unstake": self._failure_unstake,

                # Item count
                "total": len(self._targets),
                "success": len(self._success_targets),
                "failure": len(self._failure_targets),

                # Item list
                "success_targets": [target.to_dict() for target in self._success_targets],
                "failure_targets": [target.to_dict() for target in self._failure_targets],
            }

            with open(path, "w") as f:
                text = json.dumps(report, indent=4)
                f.write(text)
        except BaseException as e:
            Logger.exception(tag=TAG, msg=str(e))

        Logger.info(tag=TAG, msg=f"UnstakePatcher.write_result() end")

    @classmethod
    def from_path(cls, path: str) -> UnstakePatcher:
        data: Dict[str, Any] = cls._load(path)
        targets: List[Target] = [Target.from_dict(i) for i in data["targets"]]

        return cls(targets)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .address import Address


class Message(object):
    """Data which is sent to receipt through icx_sendTransaction json-rpc api
    """

    def __init__(self, sender: Optional['Address']=None, value: int=0) -> None:
        """Constructor

        :param sender: sender of the message (current call)
        :param value: number of loop sent with the message
        """
        # sender of the message (current call)
        self.sender = sender
        # number of loop sent with the message
        self.value = value



# Copyright 2020 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import Enum, auto
from typing import Union, Iterable, Iterator, Optional, Tuple, List

from .context.context import ContextGetter
from ..base.address import Address
from ..base.exception import (
    AccessDeniedException,
)
from ..database.db import (
    ContextDatabase,
    DatabaseObserver,
)
from ..icon_constant import Revision
from ..utils import bytes_to_hex
from ..utils.rlp import rlp_encode_bytes


class ContainerTag(Enum):
    """Container Type used in SCORE
    """
    ARRAY = b"\x00"
    DICT = b"\x01"
    VAR = b"\x02"


class KeyType(Enum):
    ARRAY = auto()
    DICT = auto()
    VAR = auto()
    SUB = auto()

    ARRAY_SIZE = auto()
    DUMMY = auto()
    LAST = auto()


class Key(object):
    def __init__(self, value: bytes, _type: KeyType = KeyType.LAST):
        self._value = value
        self._type = _type

    @property
    def value(self) -> bytes:
        return self._value

    @property
    def type(self) -> KeyType:
        return self._type

    def __str__(self):
        return f"Key(value={bytes_to_hex(self._value)}, type={self._type.name})"


def verify_single_prefix(prefix: Union[bytes, Key]) -> bool:
    if isinstance(prefix, Key):
        return True

    if isinstance(prefix, bytes) and len(prefix) > 0:
        return True

    return False


def _to_key(key: Union[bytes, Key]) -> Key:
    return Key(key) if isinstance(key, bytes) else key


def to_tag(key_type: KeyType) -> Optional[ContainerTag]:
    if KeyType.ARRAY == key_type:
        return ContainerTag.ARRAY
    elif KeyType.DICT == key_type:
        return ContainerTag.DICT
    elif KeyType.VAR == key_type:
        return ContainerTag.VAR

    return None


def _is_container_name(key: Key) -> bool:
    return key.type in {KeyType.ARRAY, KeyType.DICT, KeyType.VAR}


def to_key_body(address: Address, final_key: bytes, use_rlp: bool) -> bytes:
    """Returns the rest of a given final key excluding score_address prefix

    :param address:
    :param final_key:
    :param use_rlp:
    :return:
    """
    prefix_size = len(address.to_bytes())
    if not use_rlp:
        # Additionally remove b"|" from final_key
        prefix_size += 1

    return final_key[prefix_size:]


class PrefixStorage(object):
    """Stores multiple prefixes

    Prefix keys are some parts of final key which is used for querying stateDB

    """
    def __init__(self, keys: Iterator[Key] = None):
        self._tag: Optional[ContainerTag] = None
        self._keys = [] if keys is None else [key for key in keys]

        if len(self._keys) > 0:
            key = self._keys[-1]
            if _is_container_name(key):
                self._tag = to_tag(key.type)

    def __len__(self):
        return len(self._keys)

    def __iter__(self):
        return iter(self._keys)

    def append(self, key: Union[bytes, Key]):
        key = _to_key(key)

        if _is_container_name(key):
            self._tag = to_tag(key.type)

        self._keys.append(key)

    def get_final_key(self, key: Union[bytes, Key], use_rlp: bool) -> bytes:
        key: Key = _to_key(key)

        if use_rlp:
            return self._get_final_key_with_rlp(key)
        else:
            return self._get_final_key_with_pipe(key)

    def _get_final_key_with_pipe(self, last_key: Key) -> bytes:
        """Generate a final key with '|' separator

        :param last_key:
        :return:
        """
        if isinstance(self._tag, ContainerTag):
            keys: List[bytes] = self._get_container_final_key_with_pipe(last_key)
        else:
            keys: List[bytes] = [key.value for key in self._keys]
            keys.append(last_key.value)

        return b"|".join(keys)

    def _get_container_final_key_with_pipe(self, last_key: Key) -> List[bytes]:
        keys = []

        for key in self._keys:
            if key.type in (KeyType.ARRAY, KeyType.DICT, KeyType.VAR):
                keys.append(self._tag.value)
            keys.append(key.value)

        if last_key.type != KeyType.DUMMY:
            keys.append(last_key.value)

        return keys

    def _get_final_key_with_rlp(self, last_key: Key) -> bytes:
        """Generate final key with rlp

        :param last_key:
        :return:
        """
        def func():
            if isinstance(self._tag, ContainerTag):
                yield self._tag.value

            for _key in self._keys:
                yield rlp_encode_bytes(_key.value)

            if last_key.type != KeyType.ARRAY_SIZE and len(last_key.value) > 0:
                yield rlp_encode_bytes(last_key.value)

        return b"".join(func())

    def copy(self) -> PrefixStorage:
        return PrefixStorage(self._keys)


class KeyValuePair:
    def __init__(self, key: Optional[bytes], value: Optional[bytes], use_rlp: bool = False):
        self._key = key
        self._value = value
        self._use_rlp = use_rlp

    @property
    def key(self) -> Optional[bytes]:
        return self._key

    @property
    def value(self) -> Optional[bytes]:
        return self._value

    @property
    def use_rlp(self) -> bool:
        return self._use_rlp


class IconScoreDatabase(ContextGetter):
    """It is used in IconScore

    IconScore can access its states only through IconScoreDatabase
    """

    def __init__(
            self,
            address: 'Address',
            context_db: 'ContextDatabase',
            prev_prefixes: Iterable[Key] = None,
            prefix: Union[bytes, Key] = None,
            origin_score_db: Optional[IconScoreDatabase] = None,
    ):
        """Constructor

        :param address: the address of SCORE which this db is assigned to
        :param context_db: ContextDatabase
        """
        self._address = address
        self._context_db = context_db
        self._observer: Optional[DatabaseObserver] = None
        self._origin_score_db = origin_score_db

        self._prefixes = PrefixStorage(prev_prefixes)
        if prefix:
            self._prefixes.append(prefix)

    @property
    def address(self) -> Address:
        return self._address

    def _use_rlp(self) -> bool:
        return self._context.revision >= Revision.USE_RLP.value

    def _context_db_get(self, key: bytes) -> Optional[bytes]:
        return self._context_db.get(self._context, key)

    def _context_db_put(self, key: bytes, value: bytes):
        self._context_db.put(self._context, key, value)

    def _context_db_delete(self, key: Optional[bytes]):
        if key:
            self._context_db.delete(self._context, key)

    def get(self, key: Union[bytes, Key]) -> bytes:
        """
        Gets the value for the specified key

        :param key: key to retrieve
        :return: value for the specified key, or None if not found
        """
        key = _to_key(key)

        old_kv_pair, new_kv_pair = self._get(key)
        final_key: bytes = new_kv_pair.key if new_kv_pair.key else old_kv_pair.key
        value: bytes = new_kv_pair.value if new_kv_pair.value else old_kv_pair.value

        observer = self.__get_observer()
        if observer:
            observer.on_get(self._context, self._to_key_body(final_key), value)

        return value

    def _get(
            self,
            key: Union[bytes, Key]
    ) -> Tuple[KeyValuePair, KeyValuePair]:
        old_value, new_value = None, None
        old_final_key = self._get_final_key(key, use_rlp=False)
        new_final_key = None

        if self._use_rlp():
            new_final_key = self._get_final_key(key, use_rlp=True)
            new_value: Optional[bytes] = self._context_db_get(new_final_key)

        if new_value is None:
            old_value: Optional[bytes] = self._context_db_get(old_final_key)

        return (
            KeyValuePair(old_final_key, old_value, use_rlp=False),
            KeyValuePair(new_final_key, new_value, use_rlp=True),
        )

    def put(self, key: Union[bytes, Key], value: bytes):
        """
        Sets a value for the specified key.

        :param key: key to set
        :param value: value to set
        """
        self._validate_ownership()

        old_kv_pair, new_kv_pair = self._get(key)
        final_key: bytes = new_kv_pair.key if new_kv_pair.key else old_kv_pair.key
        prev_value: bytes = new_kv_pair.value if new_kv_pair.value else old_kv_pair.value

        observer = self.__get_observer()
        if observer:
            key_body: bytes = self._to_key_body(final_key)
            if value:
                observer.on_put(self._context, key_body, prev_value, value)
            elif prev_value:
                # If new value is None, then deletes the field
                observer.on_delete(self._context, key_body, prev_value)

        self._context_db.put(self._context, final_key, value)

    def get_sub_db(self, prefix: Union[bytes, Key]) -> IconScoreDatabase:
        """
        Returns sub db with a prefix

        :param prefix: The prefix used by this sub db.
        :return: sub db
        """
        origin_score_db = self._origin_score_db if self._origin_score_db else self

        score_db = IconScoreDatabase(
            address=self._address,
            context_db=self._context_db,
            prev_prefixes=self._prefixes,
            prefix=prefix,
            origin_score_db=origin_score_db
        )
        score_db.set_observer(self._observer)

        return score_db

    def delete(self, key: Union[bytes, Key]):
        """
        Deletes the key/value pair for the specified key.

        :param key: key to delete
        """
        self._validate_ownership()

        old_kv_pair, new_kv_pair = self._get(key)
        final_key: bytes = new_kv_pair.key if new_kv_pair.key else old_kv_pair.key
        value: bytes = new_kv_pair.value if new_kv_pair.value else old_kv_pair.value

        observer = self.__get_observer()
        if observer:
            if value:
                observer.on_delete(self._context, self._to_key_body(final_key), value)

        self._context_db_delete(old_kv_pair.key)
        self._context_db_delete(new_kv_pair.key)

    def close(self):
        self._context_db.close(self._context)

    def set_observer(self, observer: DatabaseObserver):
        self._observer = observer

    def __get_observer(self) -> Optional[DatabaseObserver]:
        if self._observer:
            return self._observer

        if self._origin_score_db:
            return self._origin_score_db.__get_observer()

        return None

    def _get_final_key(self, key: Union[bytes, Key], use_rlp: bool) -> bytes:
        """
        :params key: key passed by SCORE
        :return: key bytes
        """
        body: bytes = self._prefixes.get_final_key(key, use_rlp)
        sep = b"" if use_rlp else b"|"

        return sep.join((self._address.to_bytes(), body))

    def _validate_ownership(self):
        """Prevent a SCORE from accessing the database of another SCORE

        """
        if self._context.current_address != self.address:
            raise AccessDeniedException("Invalid database ownership")

    def _to_key_body(self, final_key: bytes) -> bytes:
        return to_key_body(self._address, final_key, self._use_rlp())



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import TYPE_CHECKING

from ..base.exception import InvalidParamsException
from ..utils.msgpack_for_db import MsgPackForDB

if TYPE_CHECKING:
    from ..base.address import Address

_TAG = "PRepAddressConverter"


class PRepAddressConverter:
    """
    Convert prep address being used for consensus in loopchain (i.e. node address) to
    P-Rep address being used for managing P-Rep (i.e. prep address).

    Loopchain distinguishes P-Reps using node address, but iconservice uses an prep address as a identification.
    """

    def __init__(self,
                 prev_node_address_mapper: dict = None,
                 node_address_mapper: dict = None):
        # For P-Rep vote reward
        if prev_node_address_mapper:
            self._prev_node_address_mapper: dict = prev_node_address_mapper
        else:
            self._prev_node_address_mapper: dict = {}

        # For assign node address validating
        if node_address_mapper:
            self._node_address_mapper: dict = node_address_mapper
        else:
            self._node_address_mapper: dict = {}

    def __str__(self):
        return f"prev_node_address_mapper={self._prev_node_address_mapper} \n" \
               f"node_address_mapper={self._node_address_mapper}"

    def to_bytes(self) -> bytes:
        version: int = 0
        return MsgPackForDB.dumps([version, self._prev_node_address_mapper])

    @classmethod
    def from_bytes(cls, data: bytes) -> 'PRepAddressConverter':
        if data is None:
            return PRepAddressConverter()

        data: list = MsgPackForDB.loads(data)
        _version = data[0]

        prev_node_address_mapper: dict = data[1]
        return PRepAddressConverter(prev_node_address_mapper=prev_node_address_mapper)

    def add_node_address(self, node: 'Address', prep: 'Address'):
        if node in self._node_address_mapper:
            raise InvalidParamsException(f"nodeAddress already in use: {node}")
        self._node_address_mapper[node] = prep

    def delete_node_address(self, node: 'Address', prep: 'Address'):
        self._add_prev_node_address(node=node, prep=prep)
        self._delete_node_address(node)

    def _delete_node_address(self, node: 'Address'):
        if node in self._node_address_mapper:
            del self._node_address_mapper[node]

    def _add_prev_node_address(self, node: 'Address', prep: 'Address'):
        if prep not in self._prev_node_address_mapper.values():
            self._prev_node_address_mapper[node] = prep

    def replace_node_address(self, node: 'Address', prep: 'Address', prev_node: 'Address'):
        self._add_prev_node_address(node=prev_node, prep=prep)
        self._delete_node_address(node=prev_node)
        self.add_node_address(node=node, prep=prep)

    def copy(self) -> 'PRepAddressConverter':
        return PRepAddressConverter(prev_node_address_mapper=copy.copy(self._prev_node_address_mapper),
                                    node_address_mapper=copy.copy(self._node_address_mapper))

    def reset_prev_node_address(self):
        self._prev_node_address_mapper.clear()

    def validate_node_address(self,
                              node: 'Address'):

        if node in self._node_address_mapper:
            raise InvalidParamsException(f"nodeAddress already in use: {node}")

    def get_prep_address_from_node_address(self,
                                           node_address: 'Address') -> 'Address':
        ret: 'Address' = self._node_address_mapper.get(node_address)
        if ret is None:
            ret = self._prev_node_address_mapper.get(node_address, node_address)
        return ret



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional, TYPE_CHECKING, Any, Dict, List

from .container import Container, ValueConverter
from .. import Address
from ..base.ComponentBase import EngineBase
from ..base.address import GOVERNANCE_SCORE_ADDRESS
from ..base.exception import ScoreNotFoundException
from ..icon_constant import IconNetworkValueType, IconServiceFlag
from ..iconscore.context.context import ContextContainer
from ..iconscore.icon_score_context_util import IconScoreContextUtil
from ..iconscore.icon_score_result import TransactionResult

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext
    from ..builtin_scores.governance.governance import Governance
    from ..precommit_data_manager import PrecommitData


class Engine(EngineBase, ContextContainer):
    TAG = "INV"

    def __init__(self):
        super().__init__()
        self._inv_container: Optional['Container'] = None

        # Warning: This mapper must be used only before migration. Do not modify.
        # 'gs' means governance SCORE
        self._get_gs_data_mapper: dict = {
            IconNetworkValueType.SERVICE_CONFIG: self._get_service_flag,
            IconNetworkValueType.STEP_PRICE: self._get_step_price_from_governance,
            IconNetworkValueType.STEP_COSTS: self._get_step_costs_from_governance,
            IconNetworkValueType.MAX_STEP_LIMITS: self._get_step_max_limits_from_governance,
            IconNetworkValueType.REVISION_CODE: self._get_revision_code_from_governance,
            IconNetworkValueType.REVISION_NAME: self._get_revision_name_from_governance,
            IconNetworkValueType.SCORE_BLACK_LIST: self._get_score_black_list,
            IconNetworkValueType.IMPORT_WHITE_LIST: self._get_import_whitelist
        }

    @property
    def inv_container(self) -> 'Container':
        return self._inv_container

    def load_inv_container(self, context: 'IconScoreContext'):
        container: Optional['Container'] = context.storage.inv.get_container(context)
        if container is None:
            # before GS data migration. load data from GS
            container: 'Container' = Container(is_migrated=False)
            self._sync_inv_container_with_governance(context, container)
        self._inv_container = container

    def update_inv_container_by_result(self,
                                       context: 'IconScoreContext',
                                       tx_result: 'TransactionResult'):
        if tx_result.status == TransactionResult.SUCCESS:
            if context.inv_container.is_migrated:
                context.inv_container.update_batch()
            elif tx_result.to == GOVERNANCE_SCORE_ADDRESS:  # only for GS update TX
                context.inv_container.update_migration_if_succeed()
                self._sync_inv_container_with_governance(context, context.inv_container)
        context.inv_container.clear_batch()

    def _sync_inv_container_with_governance(self,
                                            context: 'IconScoreContext',
                                            container: 'Container'):
        """
        Syncronize ICON Network value.
        :param context:
        :param container:
        :return:
        """
        if container.is_migrated:
            return

        try:
            self._push_context(context)
            governance = self._get_governance(context)
            for type_ in IconNetworkValueType.gs_migration_type_list():
                value: Any = self._get_gs_data_mapper[type_](context, governance)
                container.set_inv(ValueConverter.convert_for_icon_service(type_, value))
        except ScoreNotFoundException:
            pass
        finally:
            self._pop_context()

    @classmethod
    def _get_governance(cls, context: 'IconScoreContext') -> 'Governance':
        governance = \
            IconScoreContextUtil.get_icon_score(context, GOVERNANCE_SCORE_ADDRESS)
        if governance is None:
            raise ScoreNotFoundException('Governance SCORE not found')
        return governance

    @classmethod
    def _get_step_price_from_governance(cls, context: 'IconScoreContext', governance: 'Governance') -> int:
        step_price = 0
        # Gets the step price if the fee flag is on
        if IconScoreContextUtil.is_service_flag_on(context, IconServiceFlag.FEE):
            step_price = governance.getStepPrice()
        return step_price

    @classmethod
    def _get_step_costs_from_governance(cls, _, governance: 'Governance') -> Dict[str, int]:
        return governance.getStepCosts()

    @classmethod
    def _get_step_max_limits_from_governance(cls, _, governance: 'Governance') -> Dict[str, int]:
        # Gets the max step limit
        return {"invoke": governance.getMaxStepLimit("invoke"),
                "query": governance.getMaxStepLimit("query")}

    @classmethod
    def _get_service_flag(cls, context: 'IconScoreContext', governance: 'Governance') -> int:
        service_config = context.icon_service_flag
        try:
            service_config = governance.service_config
        except AttributeError:
            pass
        return service_config

    @classmethod
    def _get_revision_name_from_governance(cls, _, governance: 'Governance') -> str:
        # There is no use of revision name
        revision_name: str = ""
        if hasattr(governance, 'getRevision'):
            revision_name: str = governance.getRevision()['name']
        return revision_name

    @classmethod
    def _get_revision_code_from_governance(cls, _, governance: 'Governance') -> int:
        # Check if revision has been changed by comparing with INV engine's ICON Network value
        revision: int = 0
        if hasattr(governance, 'revision_code'):
            revision: int = governance.revision_code
        return revision

    @classmethod
    def _get_import_whitelist(cls, _, governance: 'Governance') -> Dict[str, list]:
        if hasattr(governance, 'import_white_list_cache'):
            return governance.import_white_list_cache

        return {"iconservice": ['*']}

    @classmethod
    def _get_score_black_list(cls, _, governance: 'Governance') -> List['Address']:
        score_black_list = []
        if hasattr(governance, '_score_black_list'):
            score_black_list = [address for address in governance._score_black_list]
        return score_black_list

    def commit(self, _context: 'IconScoreContext', precommit_data: 'PrecommitData'):
        # Set updated INVContainer
        self._inv_container: 'Container' = precommit_data.inv_container



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum
from typing import Tuple, Any, Union

import msgpack

from . import int_to_bytes, bytes_to_int
from ..base.address import Address
from ..base.exception import InvalidParamsException
from ..icon_constant import CHARSET_ENCODING


class TypeTag(IntEnum):
    NIL = 0
    DICT = 1
    LIST = 2
    BYTES = 3
    STRING = 4

    CUSTOM = 10
    INT = CUSTOM + 1
    ADDRESS = CUSTOM


class BaseCodec(object):

    def encode(self, obj: Any) -> Tuple[int, bytes]:
        if isinstance(obj, Address):
            return TypeTag.ADDRESS, obj.to_bytes_including_prefix()
        raise InvalidParamsException(f"Invalid encode type: {type(obj)}")

    def decode(self, t: int, b: bytes) -> Any:
        if t == TypeTag.ADDRESS:
            return Address.from_bytes_including_prefix(b)
        else:
            raise InvalidParamsException(f"UnknownType: {type(t)}")


class MsgPackForIpc(object):
    codec: 'BaseCodec' = BaseCodec()

    @classmethod
    def decode(cls, tag: int, val: bytes) -> 'Any':
        if tag == TypeTag.NIL:
            if val == b'':
                return None
            else:
                raise InvalidParamsException(f"Invalid tag type:{tag} value: {val}")
        if tag == TypeTag.BYTES:
            return val
        elif tag == TypeTag.STRING:
            return val.decode(CHARSET_ENCODING)
        elif tag == TypeTag.INT:
            return bytes_to_int(val)
        else:
            return cls.codec.decode(tag, val)

    @classmethod
    def encode(cls, o: Any) -> bytes:
        if o is None:
            return bytes([])
        if isinstance(o, int):
            return int_to_bytes(o)
        elif isinstance(o, str):
            return o.encode(CHARSET_ENCODING)
        elif isinstance(o, bytes):
            return o
        elif isinstance(o, bool):
            return b'\x01' if o else b'\x00'
        else:
            t, v = cls.codec.encode(o)
            return v

    @classmethod
    def decode_any(cls, to: Tuple) -> Any:
        tag: int = to[0]
        val: Union[bytes, dict, list] = to[1]
        if tag == TypeTag.NIL:
            return None
        elif tag == TypeTag.DICT:
            obj = {}
            for k, v in val.items():
                if isinstance(k, bytes):
                    k = k.decode(CHARSET_ENCODING)
                obj[k] = cls.decode_any(v)
            return obj
        elif tag == TypeTag.LIST:
            obj = []
            for v in val:
                obj.append(cls.decode_any(v))
            return obj
        else:
            return cls.decode(tag, val)

    @classmethod
    def encode_any(cls, o: Any) -> Tuple[int, Any]:
        if o is None:
            return TypeTag.NIL, b''
        elif isinstance(o, dict):
            m = {}
            for k, v in o.items():
                m[k] = cls.encode_any(v)
            return TypeTag.DICT, m
        elif isinstance(o, (list, tuple)):
            lst = []
            for v in o:
                lst.append(cls.encode_any(v))
            return TypeTag.LIST, lst
        elif isinstance(o, bytes):
            return TypeTag.BYTES, o
        elif isinstance(o, str):
            return TypeTag.STRING, o.encode(CHARSET_ENCODING)
        elif isinstance(o, int):
            return TypeTag.INT, int_to_bytes(o)
        else:
            return cls.codec.encode(o)

    @classmethod
    def dumps(cls, data: Any) -> bytes:
        return msgpack.dumps(data, use_bin_type=False)

    @classmethod
    def loads(cls, data: bytes) -> list:
        return msgpack.loads(data, raw=True, strict_map_key=False)



# -*- coding: utf-8 -*-

# Copyright 2020 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SystemScore module
"""

from inspect import currentframe
from typing import TYPE_CHECKING, List

from iconcommons.logger import Logger
from typing_extensions import TypedDict

from .icon_score_base import IconScoreBase, interface, external, payable, eventlog, revert
from .icon_score_base2 import InterfaceScore
from ..base.address import Address
from ..base.exception import *
from ..icon_constant import ISCORE_EXCHANGE_RATE, Revision
from ..utils import to_camel_case

if TYPE_CHECKING:
    from ..database.db import IconScoreDatabase
    from ..iconscore.icon_score_context import IconScoreContext
    from ..iiss.storage import RewardRate


class Delegation(TypedDict):
    address: Address
    value: int


class SystemScore(IconScoreBase):
    @eventlog
    def IScoreClaimed(self, iscore: int, icx: int):
        pass

    @eventlog(indexed=1)
    def IScoreClaimedV2(self, address: Address, iscore: int, icx: int):
        pass

    @eventlog
    def PRepRegistered(self, address: Address):
        pass

    @eventlog
    def PRepUnregistered(self, address: Address):
        pass

    @eventlog
    def PRepSet(self, address: Address):
        pass

    @eventlog(indexed=1)
    def ICXBurnedV2(self, address: Address, amount: int, totalSupply: int):
        pass

    def __init__(self, db: 'IconScoreDatabase') -> None:
        super().__init__(db)

    def on_install(self, **kwargs) -> None:
        super().on_install()

    def on_update(self, **kwargs) -> None:
        super().on_update()

    @external
    def setStake(self, value: int = 0) -> None:
        self._context.engine.iiss.invoke(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getStake(self, address: Address) -> dict:
        return self._context.engine.iiss.query(*self._get_params(locals_params=locals()))

    @external
    def setDelegation(self, delegations: List[Delegation] = None) -> None:
        self._context.engine.iiss.invoke(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getDelegation(self, address: Address) -> dict:
        return self._context.engine.iiss.query(*self._get_params(locals_params=locals()))

    @external
    def claimIScore(self) -> None:
        self._context.engine.iiss.invoke(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def queryIScore(self, address: Address) -> dict:
        return self._context.engine.iiss.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def estimateUnstakeLockPeriod(self) -> dict:
        return self._context.engine.iiss.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getIISSInfo(self) -> dict:
        return _handle_get_iiss_info(self._context)

    @payable
    @external
    def registerPRep(self, name: str, country: str, city: str, email: str, website: str,
                     details: str, p2pEndpoint: str, nodeAddress: "Address" = None):
        self._context.engine.prep.invoke(*self._get_params(locals_params=locals()))

    @external
    def unregisterPRep(self):
        self._context.engine.prep.invoke(*self._get_params(locals_params=locals()))

    @external
    def setPRep(self, name: str = None, country: str = None, city: str = None, email: str = None,
                website: str = None, details: str = None, p2pEndpoint: str = None, nodeAddress: "Address" = None):
        self._context.engine.prep.invoke(*self._get_params(locals_params=locals()))

    @external
    def setGovernanceVariables(self, irep: int):
        self._context.engine.prep.invoke(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getPRep(self, address: Address) -> dict:
        return self._context.engine.prep.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getPReps(self, startRanking: int = None, endRanking: int = None) -> list:
        return self._context.engine.prep.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getMainPReps(self) -> dict:
        return self._context.engine.prep.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getSubPReps(self) -> dict:
        return self._context.engine.prep.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getPRepTerm(self) -> dict:
        return self._context.engine.prep.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getInactivePReps(self) -> dict:
        return self._context.engine.prep.query(*self._get_params(locals_params=locals()))

    @external(readonly=True)
    def getScoreDepositInfo(self, address: Address) -> dict:
        # TODO check with dbtools
        # if str(from_) not in _AUTHORIZED_ACCOUNTS:
        #     raise AccessDeniedException("No permission")

        context = self._context
        deposit_info = context.engine.fee.get_deposit_info(
            context, address, context.block.height
        )
        return None if deposit_info is None else deposit_info.to_dict(to_camel_case)

    @payable
    @external
    def burn(self):
        context = self._context
        if context.revision < Revision.BURN_V2_ENABLED.value:
            revert("burn is not enabled")

        context.engine.issue.burn(context, self.msg.sender, self.msg.value)

    def _get_params(self, locals_params: dict) -> tuple:
        method = currentframe().f_back.f_code.co_name
        params: dict = self._del_self_in_params(locals_params)
        return self._context, method, params

    def _del_self_in_params(self, kw_args: dict) -> dict:
        params = dict(kw_args)
        del params["self"]

        return params


class InterfaceSystemScore(InterfaceScore):
    @interface
    def setStake(self, value: int) -> None: pass

    @interface
    def getStake(self, address: Address) -> dict: pass

    @interface
    def estimateUnstakeLockPeriod(self) -> dict: pass

    @interface
    def setDelegation(self, delegations: List[Delegation] = None): pass

    @interface
    def getDelegation(self, address: Address) -> dict: pass

    @interface
    def claimIScore(self): pass

    @interface
    def queryIScore(self, address: Address) -> dict: pass

    @interface
    def getIISSInfo(self) -> dict: pass

    @interface
    def getPRep(self, address: Address) -> dict: pass

    @interface
    def getPReps(self, startRanking: int, endRanking: int) -> list: pass

    @interface
    def getMainPReps(self) -> dict: pass

    @interface
    def getSubPReps(self) -> dict: pass

    @interface
    def getPRepTerm(self) -> dict: pass

    @interface
    def getInactivePReps(self) -> dict: pass

    @interface
    def getScoreDepositInfo(self, address: Address) -> dict: pass

    @interface
    def burn(self): pass


def _create_rc_result(context: "IconScoreContext", start_block: int, end_block: int) -> dict:
    rc_result = dict()
    if start_block < 0 or end_block < 0:
        return rc_result

    iscore, request_block_height, rc_state_hash = \
        context.storage.rc.get_calc_response_from_rc()
    if iscore == -1:
        return rc_result

    if request_block_height != end_block:
        Logger.warning(
            tag="ISE",
            msg=f"Response block height is not matched to the request: "
            f"response block height:{request_block_height} "
            f"request block height:{end_block}",
        )
        return rc_result

    rc_result["iscore"] = iscore
    rc_result["estimatedICX"] = iscore // ISCORE_EXCHANGE_RATE
    rc_result["startBlockHeight"] = start_block
    rc_result["endBlockHeight"] = end_block
    rc_result["stateHash"] = rc_state_hash

    return rc_result


def _handle_get_iiss_info(context: "IconScoreContext") -> dict:
    term = context.engine.prep.term
    reward_rate: "RewardRate" = context.storage.iiss.get_reward_rate(context)

    response = {
        "blockHeight": context.block.height,
        "variable": {
            "irep": term.irep if term else 0,
            "rrep": reward_rate.reward_prep
        },
    }

    calc_start_block, calc_end_block = context.storage.meta.get_last_calc_info(context)

    next_calculation: int = calc_end_block
    if calc_start_block < 0 or context.block.height != next_calculation:
        next_calculation: Optional[
            int
        ] = context.storage.iiss.get_end_block_height_of_calc(context)
        if next_calculation is None:
            next_calculation = -1
    response["nextCalculation"] = next_calculation + 1

    term_start_block, term_end_block = context.storage.meta.get_last_term_info(context)

    if term_end_block < 0 or context.block.height != term_end_block:
        term_end_block: int = term.end_block_height if term else -1
    response["nextPRepTerm"] = term_end_block + 1

    response["rcResult"] = _create_rc_result(context, calc_start_block, calc_end_block)

    return response



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = "get_backup_filename"

import os


def get_backup_filename(block_height: int) -> str:
    """

    :param block_height: the height of the block where we want to go back
    :return:
    """
    return f"{block_height:010d}.bak"


def check_backup_exists(backup_root_path: str, current_block_height: int, rollback_block_height: int) -> bool:
    """Check if backup files for rollback exist

    :param backup_root_path: the directory where backup files are located
    :param current_block_height: current state before rollback
    :param rollback_block_height: final state after rollback
    :return: True(exist) False(not exist)
    """
    if current_block_height < 1 or \
            rollback_block_height < 0 or \
            rollback_block_height > current_block_height:
        return False

    for block_height in range(current_block_height - 1, rollback_block_height - 1, -1):
        filename = get_backup_filename(block_height)
        path = os.path.join(backup_root_path, filename)
        if not os.path.isfile(path):
            return False

    return True



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""functions and classes to handle address
"""

import hashlib
from enum import IntEnum
from typing import Optional

from .exception import InvalidParamsException
from ..icon_constant import DATA_BYTE_ORDER, ICON_DEX_DB_NAME, GOVERNANCE_ADDRESS, SYSTEM_ADDRESS
from ..utils import is_lowercase_hex_string, int_to_bytes

ICON_EOA_ADDRESS_PREFIX = 'hx'
ICON_CONTRACT_ADDRESS_PREFIX = 'cx'
ICON_EOA_ADDRESS_BYTES_SIZE = 20
ICON_CONTRACT_ADDRESS_BYTES_SIZE = 21
ICON_ADDRESS_BYTES_SIZE = 21
ICON_ADDRESS_BODY_SIZE = 20


def is_icon_address_valid(address: str) -> bool:
    """Check whether address is in icon address format or not

    :param address: (str) address string including prefix
    :return: (bool)
    """
    try:
        if isinstance(address, str) and len(address) == 42:
            prefix, body = split_icon_address(address)
            if prefix == ICON_EOA_ADDRESS_PREFIX or \
                    prefix == ICON_CONTRACT_ADDRESS_PREFIX:
                return is_lowercase_hex_string(body)
    finally:
        pass

    return False


def split_icon_address(address: str) -> (str, str):
    """Split icon address into 2-char prefix and 40-char address body

    :param address: 42-char address string
    :return: prefix, body
    """
    return address[:2], address[2:]


class AddressPrefix(IntEnum):
    """
    Enumeration of Address prefix

    - CONTRACT: Contract Account
    - EOA: Externally Owned Account


    """

    EOA = 0

    CONTRACT = 1

    def __str__(self) -> str:
        if self == AddressPrefix.EOA:
            return ICON_EOA_ADDRESS_PREFIX
        if self == AddressPrefix.CONTRACT:
            return ICON_CONTRACT_ADDRESS_PREFIX

    @staticmethod
    def from_string(prefix: str):
        """
        Returns address prefix enumerator

        :param prefix: 2-byte address prefix (hx or cx)
        :return: (AddressPrefix) address prefix enumerator
        """
        if prefix == ICON_EOA_ADDRESS_PREFIX:
            return AddressPrefix.EOA
        if prefix == ICON_CONTRACT_ADDRESS_PREFIX:
            return AddressPrefix.CONTRACT

        raise InvalidParamsException('Invalid address prefix')


class Address(object):
    """Address class
    """

    def __init__(self,
                 address_prefix: AddressPrefix,
                 address_body: bytes, ignore_length_validate: bool = False) -> None:
        """Constructor

        :param address_prefix: address prefix enumerator
        :param address_body: 20-byte address body
        """

        if not isinstance(address_prefix, AddressPrefix):
            raise InvalidParamsException('Invalid address prefix type')
        if not isinstance(address_body, bytes):
            raise InvalidParamsException('Invalid address body type')

        if not ignore_length_validate:
            if len(address_body) != ICON_ADDRESS_BODY_SIZE:
                raise InvalidParamsException('Address length is not 20 in bytes')

        self.__prefix = address_prefix
        self.__body = address_body

    @property
    def prefix(self) -> AddressPrefix:
        """Returns address prefix part

        :return: :class:`.AddressPrefix` AddressPrefix.EOA(0) or AddressPrefix.CONTRACT(1)
        """
        return self.__prefix

    @property
    def body(self) -> bytes:
        """Returns 20-byte address body part

        :return: 20 byte data standing for address
        """
        return self.__body

    def __eq__(self, other) -> bool:
        """operator == overriding

        :return: bool
        """
        return \
            isinstance(other, Address) \
            and self.__prefix == other.prefix \
            and self.__body == other.body

    def __ne__(self, other) -> bool:
        """operator != overriding

        :return: (bool)
        """
        return not self.__eq__(other)

    def __str__(self) -> str:
        """operator str() overriding

        returns prefix(2) + 40-char hexadecimal address

        :return: (str) 42-char address
        """
        return f'{str(self.prefix)}{self.body.hex()}'

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self) -> int:
        """Returns a hash value for this object

        :return: hash value
        """
        return hash(self.__prefix.to_bytes(1, DATA_BYTE_ORDER) + self.__body)

    @property
    def is_contract(self) -> bool:
        """
        Whether the address is SCORE

        :return: True(contract) False(Not contract)
        """
        return self.prefix == AddressPrefix.CONTRACT

    @staticmethod
    def from_string(address: str):
        """
        creates an address object from given 42-char string `address`

        :return: :class:`.Address`
        """

        if not is_icon_address_valid(address):
            raise InvalidParamsException('Invalid address')

        prefix, body = split_icon_address(address)

        address_prefix = AddressPrefix.from_string(prefix)
        address_body = bytes.fromhex(body)

        return Address(address_prefix, address_body)

    @staticmethod
    def from_data(prefix: AddressPrefix, data: bytes) -> Optional['Address']:
        """
        creates an address object using given body bytes

        :param prefix: address prefix
        :param data: 20-bytes address body
        :return: :class:`.Address`
        """
        try:
            hash_value = hashlib.sha3_256(data).digest()
            return Address(prefix, hash_value[-20:])
        except:
            return None

    @staticmethod
    def from_bytes(buf: bytes) -> Optional['Address']:
        """Create Address object from bytes data

        :param buf: :class:`.bytes` bytes data including Address information
        :return: :class:`.Address`
        """
        if not isinstance(buf, bytes):
            return None

        size: int = len(buf)
        if size not in (ICON_ADDRESS_BODY_SIZE, ICON_ADDRESS_BYTES_SIZE):
            return None

        if size == ICON_ADDRESS_BYTES_SIZE:
            prefix: 'AddressPrefix' = AddressPrefix(buf[0])
            return Address(prefix, buf[1:])
        else:
            return Address(AddressPrefix.EOA, buf)

    def to_bytes(self) -> bytes:
        """
        Returns data as bytes from the address object

        :return: :class:`.bytes` data including information of Address object
        """
        if self.__prefix == AddressPrefix.EOA:
            return self.__body
        else:
            return self.__prefix.to_bytes(1, DATA_BYTE_ORDER) + self.__body

    @staticmethod
    def from_bytes_including_prefix(buf: bytes) -> Optional['Address']:
        try:
            return Address(address_prefix=AddressPrefix(buf[0]), address_body=buf[1:])
        except:
            return None

    def to_bytes_including_prefix(self) -> bytes:
        return self.__prefix.to_bytes(1, DATA_BYTE_ORDER) + self.__body

    @staticmethod
    def from_prefix_and_int(prefix: 'AddressPrefix', num: int):
        num_bytes = int_to_bytes(num)
        zero_size = 20 - len(num_bytes)
        if zero_size < 0:
            raise InvalidParamsException(f'num_bytes is over 20 bytes num: {num}')
        return Address(prefix, b'\x00' * zero_size + num_bytes)


class MalformedAddress(Address):
    """This class only exists to support an invalid format address which was created by legacy bug
    """
    def __init__(self,
                 address_prefix: AddressPrefix,
                 address_body: bytes) -> None:
        """Constructor

        :param address_prefix: address prefix enumerator
        :param address_body: 20-byte address body
        """

        super().__init__(address_prefix, address_body, ignore_length_validate=True)

    @staticmethod
    def from_string(address: str):
        """Create Address object from 42-char address

        :return: :class:`.Address`
        """

        try:
            if address.startswith('hx'):
                body = address[2:]
            else:
                body = address

            address_body = bytes.fromhex(body)
        except:
            raise InvalidParamsException('Invalid address')

        return MalformedAddress(AddressPrefix.EOA, address_body)


# cx0000000000000000000000000000000000000000
SYSTEM_SCORE_ADDRESS = Address.from_string(SYSTEM_ADDRESS)
ZERO_SCORE_ADDRESS = SYSTEM_SCORE_ADDRESS
# cx0000000000000000000000000000000000000001
GOVERNANCE_SCORE_ADDRESS = Address.from_string(GOVERNANCE_ADDRESS)
ICX_ENGINE_ADDRESS = Address.from_data(AddressPrefix.CONTRACT, ICON_DEX_DB_NAME.encode())


def generate_score_address_for_tbears(score_path: str) -> 'Address':
    """

    :param score_path:
        The path of a SCORE which is under development with tbears
    :return:
    """
    project_name = score_path.split('/')[-1]
    return Address.from_data(AddressPrefix.CONTRACT, project_name.encode())


def generate_score_address(from_: 'Address',
                           timestamp: int,
                           nonce: int = None) -> 'Address':
    """Generates a SCORE address from the transaction information.

    :param from_:
    :param timestamp:
    :param nonce:
    :return: score address
    """
    data = from_.body + timestamp.to_bytes(32, DATA_BYTE_ORDER)
    if nonce:
        data += nonce.to_bytes(32, DATA_BYTE_ORDER)

    return Address.from_data(AddressPrefix.CONTRACT, data)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from .icon_score_constant import STR_FALLBACK
from .icon_score_context_util import IconScoreContextUtil
from .internal_call import InternalCall
from ..base.address import GOVERNANCE_SCORE_ADDRESS

if TYPE_CHECKING:
    from .icon_score_context import IconScoreContext
    from ..base.address import Address


class Icx(object):
    """Class for handling ICX coin transfer

    These functions are intended to be used for SCORE development.
    """

    def __init__(self, context: 'IconScoreContext', address: 'Address') -> None:
        """Constructor
        """
        self._context = context
        self._address = address

    def transfer(self, addr_to: 'Address', amount: int) -> None:
        """
        transfer the amount of icx to the given 'addr_to'
        If failed, an exception will be raised

        :param addr_to: receiver address
        :param amount: the amount of icx to transfer (unit: loop)
        """
        InternalCall.other_external_call(self._context, self._address, addr_to, amount, STR_FALLBACK)

    def send(self, addr_to: 'Address', amount: int) -> bool:
        """
        transfer the amount of icx to the given 'addr_to'

        :param addr_to: receiver address
        :param amount: the amount of icx to transfer (unit: loop)
        :return: True(success) False(failed)
        """
        try:
            self.transfer(addr_to, amount)
            if not addr_to.is_contract and self._is_icx_send_defective():
                return False
            return True
        except:
            return False

    def get_balance(self, address: 'Address') -> int:
        """
        Returns the ICX balance of given address

        :param address: address
        :return: ICX balance of given address
        """
        return InternalCall.icx_get_balance(self._context, address)

    # noinspection PyBroadException
    def _is_icx_send_defective(self) -> bool:
        try:
            governance_score = IconScoreContextUtil.get_builtin_score(
                self._context, GOVERNANCE_SCORE_ADDRESS)

            if hasattr(governance_score, 'getVersion'):
                version = governance_score.getVersion()
                return version == '0.0.2'
        except BaseException:
            pass

        return False



# -*- coding: utf-8 -*-
# Copyright 2020 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = "get_score_api"

from inspect import Signature, Parameter
from typing import List, Dict, Mapping, Iterable, Any, Union

from . import get_origin, get_args, is_struct
from .conversion import is_base_type
from .element import (
    ScoreElementMetadata,
    FunctionMetadata,
    EventLogMetadata,
    ScoreFlag,
)
from ..icon_score_constant import STR_FALLBACK
from ...base.exception import (
    IllegalFormatException,
    InvalidParamsException,
    InternalServiceErrorException,
)

"""Utils to support icx_getScoreApi method
"""


def get_score_api(elements: Iterable[ScoreElementMetadata]) -> List:
    """Returns score api used in icx_getScoreApi JSON-RPC method

    :param elements:
    :return:
    """

    api = []

    for element in elements:
        if isinstance(element, FunctionMetadata):
            func: FunctionMetadata = element
            if func.flag == ScoreFlag.PAYABLE:
                continue

            item = _get_function(func.name, func.signature, func.is_readonly, func.is_payable)
        elif isinstance(element, EventLogMetadata):
            eventlog: EventLogMetadata = element
            item = _get_eventlog(eventlog.name, eventlog.signature, eventlog.indexed_args_count)
        else:
            raise InternalServiceErrorException(f"Invalid score element: {element} {type(element)}")

        api.append(item)

    return api


def _get_function(func_name: str, sig: Signature, is_readonly: bool, is_payable: bool) -> Dict:
    if _is_fallback(func_name, sig, is_payable):
        return _get_fallback_function()
    else:
        return _get_normal_function(func_name, sig, is_readonly, is_payable)


def _get_normal_function(func_name: str, sig: Signature, is_readonly: bool, is_payable: bool) -> Dict:
    ret = {
        "name": func_name,
        "type": "function",
        "inputs": _get_inputs(sig.parameters),
        "outputs": _get_outputs(sig.return_annotation)
    }

    if is_readonly:
        ret["readonly"] = True

    if is_payable:
        ret["payable"] = True

    return ret


def _is_fallback(func_name: str, sig: Signature, is_payable: bool) -> bool:
    ret: bool = func_name == STR_FALLBACK and is_payable
    if ret:
        if len(sig.parameters) > 1:
            raise InvalidParamsException("Invalid fallback signature")

        return_annotation = sig.return_annotation
        if return_annotation not in (None, Signature.empty):
            raise InvalidParamsException("Invalid fallback signature")

    return ret


def _get_fallback_function() -> Dict:
    return {
        "name": STR_FALLBACK,
        "type": STR_FALLBACK,
        "payable": True,
    }


def _get_inputs(params: Mapping[str, Parameter]) -> list:
    inputs = []

    for name, param in params.items():
        annotation = param.annotation
        type_hint = str if annotation is Parameter.empty else annotation

        inputs.append(_get_input(name, type_hint, param.default))

    return inputs


def _get_input(name: str, type_hint: type, default: Any) -> Dict:
    inp = {"name": name}

    # Add default parameter value to score api
    if default is not Parameter.empty:
        if default is not None and not isinstance(default, type_hint):
            raise InvalidParamsException(
                f"Default params type mismatch. value: {default} type: {type_hint}")

        inp["default"] = default

    type_hints: List[type] = _split_type_hint(type_hint)
    inp["type"] = _type_hints_to_name(type_hints)

    last_type_hint: type = type_hints[-1]

    if is_struct(last_type_hint):
        inp["fields"] = _get_fields(last_type_hint)

    return inp


def _split_type_hint(type_hint: type) -> List[type]:
    type_hints = [type_hint]
    ret = []

    while len(type_hints) > 0:
        type_hint = type_hints.pop(0)
        origin: type = get_origin(type_hint)
        ret.append(origin)

        if origin is list:
            args = get_args(type_hint)
            if len(args) != 1:
                raise IllegalFormatException(f"Invalid type: {type_hint}")

            type_hints.append(args[0])
        elif origin is Union:
            args = get_args(type_hint)
            if not (len(args) == 2 and args[1] is type(None)):
                raise IllegalFormatException(f"Invalid type: {type_hint}")

            type_hints.append(args[0])

    return ret


def _type_hints_to_name(type_hints: List[type]) -> str:
    def func():
        for _type in type_hints:
            if _type is Union:
                continue

            if _type is list:
                yield "[]"
            elif is_base_type(_type):
                yield _type.__name__
            elif is_struct(_type):
                yield "struct"

    return "".join(func())


def _type_hint_to_name(type_hint: type) -> str:
    if is_base_type(type_hint):
        return type_hint.__name__
    elif is_struct(type_hint):
        return "struct"

    raise IllegalFormatException(f"Invalid type: {type_hint}")


def _get_fields(struct: type) -> List[dict]:
    """Returns fields info from struct

    :param struct: struct type
    :return:
    """
    # annotations is a dictionary containing key-type pair
    # which has field_name as a key and type as a value
    annotations = struct.__annotations__

    fields = []
    for name, type_hint in annotations.items():
        field = {"name": name}

        type_hints: List[type] = _split_type_hint(type_hint)
        field["type"] = _type_hints_to_name(type_hints)

        last_type_hint: type = type_hints[-1]
        if is_struct(last_type_hint):
            field["fields"] = _get_fields(last_type_hint)

        fields.append(field)

    return fields


def _get_outputs(type_hint: type) -> List:
    origin = get_origin(type_hint)

    if is_base_type(origin):
        type_name = origin.__name__
    elif is_struct(origin) or origin is dict:
        type_name = "dict"
    elif origin is list:
        type_name = "list"
    else:
        return []

    return [{"type": type_name}]


def _get_eventlog(func_name: str, sig: Signature, indexed_args_count: int) -> Dict:
    params = sig.parameters

    inputs = []
    for name, param in params.items():
        annotation = param.annotation
        type_hint = str if annotation is Parameter.empty else annotation
        inp: Dict = _get_input(name, type_hint, param.default)
        if len(inputs) < indexed_args_count:
            inp["indexed"] = True

        inputs.append(inp)

    return {
        "name": func_name,
        "type": "eventlog",
        "inputs": inputs
    }



# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from decimal import Decimal
from enum import IntEnum
from typing import List, Dict, Optional

from .deposit import Deposit
from .deposit_meta import DepositMeta
from ..base.ComponentBase import EngineBase
from ..base.exception import InvalidRequestException, InvalidParamsException
from ..base.type_converter import TypeConverter
from ..base.type_converter_templates import ParamType
from ..icon_constant import ICX_IN_LOOP, Revision
from ..iconscore.icon_score_event_log import EventLogEmitter

if typing.TYPE_CHECKING:
    from ..base.address import Address
    from ..deploy.storage import IconScoreDeployInfo
    from ..iconscore.icon_score_context import IconScoreContext

FIXED_TERM = True
FIXED_RATIO_PER_MONTH = '0.08'
BLOCKS_IN_ONE_MONTH = 1_296_000


class DepositInfo:
    """
    Deposit information of a SCORE
    """

    def __init__(self, score_address: 'Address'):
        # SCORE address
        self.score_address: 'Address' = score_address
        # List of deposits
        self.deposits: List[Deposit] = []
        # available virtual STEPs to use
        self.available_virtual_step: int = 0
        # available deposits to use
        self.available_deposit: int = 0

    def to_dict(self, casing: Optional = None) -> dict:
        """
        Returns properties as `dict`
        :return: a dict
        """
        new_dict = {}
        for key, value in self.__dict__.items():
            if value is None:
                # Excludes properties which have `None` value
                continue

            new_key = casing(key) if casing else key
            if isinstance(value, list):
                new_dict[new_key] = [v.to_dict(casing) for v in value if isinstance(v, Deposit)]
            else:
                new_dict[new_key] = value

        return new_dict


class Engine(EngineBase):
    """
    Presenter of the fee operation.

    [Role]
    - State DB CRUD
    - Business logic (inc. Calculation)
    """

    _MIN_DEPOSIT_AMOUNT = 5_000 * ICX_IN_LOOP
    _MAX_DEPOSIT_AMOUNT = 100_000 * ICX_IN_LOOP
    _MIN_DEPOSIT_TERM = BLOCKS_IN_ONE_MONTH
    _MAX_DEPOSIT_TERM = _MIN_DEPOSIT_TERM if FIXED_TERM else BLOCKS_IN_ONE_MONTH * 24

    def get_deposit_info(self,
                         context: 'IconScoreContext',
                         score_address: 'Address',
                         block_height: int) -> 'DepositInfo':
        """
        Gets the SCORE deposit information

        :param context: IconScoreContext
        :param score_address: SCORE address
        :param block_height: current block height
        :return: score deposit information in dict
                - SCORE Address
                - Amount of issued total virtual step
                - Amount of Used total virtual step
                - deposits in list
        """

        self._check_score_valid(context, score_address)

        deposit_meta = self._get_or_create_deposit_meta(context, score_address)

        deposit_info = DepositInfo(score_address)

        # Appends all deposits
        for deposit in self._deposit_generator(context, deposit_meta.head_id):
            deposit_info.deposits.append(deposit)

            # Retrieves available virtual STEPs and deposits
            if block_height < deposit.expires:
                deposit_info.available_virtual_step += deposit.remaining_virtual_step
                deposit_info.available_deposit += \
                    max(deposit.remaining_deposit - deposit.min_remaining_deposit, 0)

        return deposit_info if len(deposit_info.deposits) > 0 else None

    def add_deposit(self,
                    context: 'IconScoreContext',
                    tx_hash: bytes,
                    sender: 'Address',
                    score_address: 'Address',
                    amount: int,
                    block_height: int,
                    term: int):
        """
        Deposits ICXs for the SCORE.
        It may be issued the virtual STEPs for the SCORE to be able to pay share fees.

        :param context: IconScoreContext
        :param tx_hash: tx hash of the deposit transaction
        :param sender: ICX sender
        :param score_address: SCORE
        :param amount: amount of ICXs in loop
        :param block_height: current block height
        :param term: deposit term in blocks
        """
        # [Sub Task]
        # - Deposits ICX
        # - Calculates Virtual Step
        # - Updates Deposit Data

        if (FIXED_TERM and amount < self._MIN_DEPOSIT_AMOUNT) \
                or not (self._MIN_DEPOSIT_AMOUNT <= amount <= self._MAX_DEPOSIT_AMOUNT):
            raise InvalidRequestException('Invalid deposit amount')

        if not (self._MIN_DEPOSIT_TERM <= term <= self._MAX_DEPOSIT_TERM):
            raise InvalidRequestException('Invalid deposit term')

        self._check_score_ownership(context, sender, score_address)

        # Withdraws from sender's account
        sender_account = context.storage.icx.get_account(context, sender)
        sender_account.withdraw(amount)
        context.storage.icx.put_account(context, sender_account)

        deposit = Deposit(tx_hash, score_address, sender, amount)
        deposit.created = block_height
        deposit.expires = block_height + term

        step_price = context.step_counter.step_price
        deposit.virtual_step_issued = \
            VirtualStepCalculator.calculate_virtual_step(amount, term, step_price)

        self._append_deposit(context, deposit)

    def _append_deposit(self, context: 'IconScoreContext', deposit: 'Deposit'):
        """
        Append deposit data to storage
        """

        deposit_meta = self._get_or_create_deposit_meta(context, deposit.score_address)

        deposit.prev_id = deposit_meta.tail_id
        context.storage.fee.put_deposit(context, deposit)

        # Link to previous item
        if deposit.prev_id is not None:
            prev_deposit = context.storage.fee.get_deposit(context, deposit.prev_id)
            prev_deposit.next_id = deposit.id
            context.storage.fee.put_deposit(context, prev_deposit)

        # Update head info
        if deposit_meta.head_id is None:
            deposit_meta.head_id = deposit.id

        if deposit_meta.available_head_id_of_virtual_step is None:
            deposit_meta.available_head_id_of_virtual_step = deposit.id

        if deposit_meta.available_head_id_of_deposit is None:
            deposit_meta.available_head_id_of_deposit = deposit.id

        if deposit_meta.expires_of_virtual_step < deposit.expires:
            deposit_meta.expires_of_virtual_step = deposit.expires

        if deposit_meta.expires_of_deposit < deposit.expires:
            deposit_meta.expires_of_deposit = deposit.expires

        deposit_meta.tail_id = deposit.id
        context.storage.fee.put_deposit_meta(context, deposit.score_address, deposit_meta)

    def withdraw_deposit(self,
                         context: 'IconScoreContext',
                         sender: 'Address',
                         deposit_id: bytes,
                         block_height: int) -> (int, int):
        """
        Withdraws deposited ICXs from given id.
        It may be paid the penalty if the expiry has not been met.

        :param context: IconScoreContext
        :param sender: msg sender address
        :param deposit_id: deposit id, should be tx hash of deposit transaction
        :param block_height: current block height
        :return: returning amount of icx, penalty amount of icx
        """
        # [Sub Task]
        # - Checks if the contract term has expired
        # - If the term has not finished, it calculates and applies to a penalty
        # - Update ICX

        deposit = self.get_deposit(context, deposit_id)

        if deposit.sender != sender:
            raise InvalidRequestException('Invalid sender')
        if deposit.score_address != context.tx.to:
            raise InvalidRequestException('Invalid SCORE address')

        step_price = context.step_counter.step_price
        penalty = self._calculate_penalty(deposit, block_height, step_price)

        withdrawal_amount = deposit.remaining_deposit - penalty

        if withdrawal_amount < 0:
            raise InvalidRequestException("Failed to withdraw deposit")

        if penalty > 0:
            # Move the penalty amount to the treasury account
            treasury_account = context.storage.icx.get_treasury_account(context)
            treasury_account.deposit(penalty)
            context.storage.icx.put_account(context, treasury_account)

        if withdrawal_amount > 0:
            # Send the withdrawal amount of ICX to sender account
            sender_account = context.storage.icx.get_account(context, sender)
            sender_account.deposit(withdrawal_amount)
            context.storage.icx.put_account(context, sender_account)

        self._delete_deposit(context, deposit, block_height)

        return withdrawal_amount, penalty

    def _delete_deposit(self, context: 'IconScoreContext', deposit: 'Deposit', block_height: int) -> None:
        """
        Deletes deposit information from storage
        """
        # Updates the previous link
        if deposit.prev_id is not None:
            prev_deposit = context.storage.fee.get_deposit(context, deposit.prev_id)
            prev_deposit.next_id = deposit.next_id
            context.storage.fee.put_deposit(context, prev_deposit)

        # Updates the next link
        if deposit.next_id is not None:
            next_deposit = context.storage.fee.get_deposit(context, deposit.next_id)
            next_deposit.prev_id = deposit.prev_id
            context.storage.fee.put_deposit(context, next_deposit)

        # Update index info
        deposit_meta = context.storage.fee.get_deposit_meta(context, deposit.score_address)
        deposit_meta_changed = False

        if deposit_meta.head_id == deposit.id:
            deposit_meta.head_id = deposit.next_id
            deposit_meta_changed = True

        if deposit.id in (deposit_meta.available_head_id_of_virtual_step, deposit_meta.available_head_id_of_deposit):
            gen = self._deposit_generator(context, deposit.next_id)
            next_available_deposit = \
                next(filter(lambda d: block_height < d.expires, gen), None)
            next_deposit_id = \
                next_available_deposit.id if next_available_deposit is not None else None

            if deposit_meta.available_head_id_of_virtual_step == deposit.id:
                # Search for next deposit id which is available to use virtual step
                deposit_meta.available_head_id_of_virtual_step = next_deposit_id

            if deposit_meta.available_head_id_of_deposit == deposit.id:
                # Search for next deposit id which is available to use the deposited ICX
                deposit_meta.available_head_id_of_deposit = next_deposit_id

            deposit_meta_changed = True

        if deposit_meta.expires_of_virtual_step == deposit.expires:
            gen = self._deposit_generator(context, deposit_meta.available_head_id_of_virtual_step)
            max_expires = max(map(lambda d: d.expires, gen), default=-1)
            deposit_meta.expires_of_virtual_step = max_expires if max_expires > block_height else -1
            deposit_meta_changed = True

        if deposit_meta.expires_of_deposit == deposit.expires:
            gen = self._deposit_generator(context, deposit_meta.available_head_id_of_deposit)
            max_expires = max(map(lambda d: d.expires, gen), default=-1)
            deposit_meta.expires_of_deposit = max_expires if max_expires > block_height else -1
            deposit_meta_changed = True

        if deposit_meta.tail_id == deposit.id:
            deposit_meta.tail_id = deposit.prev_id
            deposit_meta_changed = True

        if deposit_meta_changed:
            # Updates if the information has been changed
            context.storage.fee.put_deposit_meta(context, deposit.score_address, deposit_meta)

        # Deletes deposit info
        context.storage.fee.delete_deposit(context, deposit.id)

    def get_deposit(self, context: 'IconScoreContext', deposit_id: bytes) -> Deposit:
        """
        Gets the deposit data.
        Raise an exception if the deposit from the given id does not exist.

        :param context: IconScoreContext
        :param deposit_id: deposit id, should be tx hash of deposit transaction
        :return: deposit data
        """

        self._check_deposit_id(deposit_id)

        deposit = context.storage.fee.get_deposit(context, deposit_id)
        if deposit is None:
            raise InvalidRequestException('Deposit not found')

        return deposit

    def check_score_available(self, context: 'IconScoreContext', score_address: 'Address', block_height: int):
        """
        Check if the SCORE is available.
        If the SCORE is sharing fee, SCORE should be able to pay the fee,
        otherwise, the SCORE is not available.

        :param context: IconScoreContext
        :param score_address: SCORE address
        :param block_height: current block height
        """
        deposit_meta: 'DepositMeta' = self._get_or_create_deposit_meta(context, score_address)

        if self._is_score_sharing_fee(deposit_meta):
            virtual_step_available = \
                block_height < deposit_meta.expires_of_virtual_step \
                and deposit_meta.available_head_id_of_virtual_step is not None

            deposit_available = \
                block_height < deposit_meta.expires_of_deposit \
                and deposit_meta.available_head_id_of_deposit is not None

            if not virtual_step_available and not deposit_available:
                raise InvalidRequestException('Out of deposit balance')

    @staticmethod
    def _is_score_sharing_fee(deposit_meta: 'DepositMeta') -> bool:
        return deposit_meta is not None and deposit_meta.head_id is not None

    def _get_fee_sharing_proportion(self, context: 'IconScoreContext', deposit_meta: 'DepositMeta'):

        if not self._is_score_sharing_fee(deposit_meta):
            # If there are no deposits, ignores the fee sharing ratio that the SCORE set.
            return 0

        return context.fee_sharing_proportion

    def charge_transaction_fee(self,
                               context: 'IconScoreContext',
                               sender: 'Address',
                               to: 'Address',
                               step_price: int,
                               used_step: int,
                               block_height: int) -> Dict['Address', int]:
        """
        Charges fees for the used STEPs.
        It can pay by shared if the msg recipient set to share fees.

        :param context: IconScoreContext
        :param sender: msg sender
        :param to: msg recipient
        :param step_price: current STEP price
        :param used_step: used STEPs
        :param block_height: current block height
        :return Address-used_step dict
        """

        recipient_step = 0

        if to.is_contract:
            recipient_step = self._charge_fee_from_score(
                context, to, step_price, used_step, block_height)

        sender_step = used_step - recipient_step
        context.engine.icx.charge_fee(context, sender, sender_step * step_price)

        step_used_details = {}

        if sender_step > 0:
            step_used_details[sender] = sender_step

        if recipient_step > 0:
            step_used_details[to] = recipient_step

        return step_used_details

    def _charge_fee_from_score(self,
                               context: 'IconScoreContext',
                               score_address: 'Address',
                               step_price: int,
                               used_step: int,
                               block_height: int) -> int:
        """
        Charges fees from SCORE
        Returns total STEPs SCORE paid
        """

        deposit_meta = context.storage.fee.get_deposit_meta(context, score_address)

        # Amount of STEPs that SCORE will pay
        required_step = used_step * self._get_fee_sharing_proportion(context, deposit_meta) // 100
        score_used_step = 0

        if required_step > 0:
            score_used_step, deposit_meta_changed = self._charge_fee_from_virtual_step(
                context, deposit_meta, required_step, block_height)

            if score_used_step < required_step:
                required_icx = (required_step - score_used_step) * step_price
                charged_icx, deposit_indices_changed = self._charge_fee_from_deposit(
                    context, deposit_meta, required_icx, block_height)

                score_used_step += charged_icx // step_price
                deposit_meta_changed: bool = deposit_meta_changed or deposit_indices_changed

            if deposit_meta_changed:
                # Updates if the information has been changed
                context.storage.fee.put_deposit_meta(context, score_address, deposit_meta)

        return score_used_step

    def _charge_fee_from_virtual_step(self,
                                      context: 'IconScoreContext',
                                      deposit_meta: 'DepositMeta',
                                      required_step: int,
                                      block_height: int) -> (int, bytes):
        """
        Charges fees from available virtual STEPs
        Returns total charged amount and whether the properties of 'deposit_meta' are changed
        """
        charged_step = 0
        should_update_expire = False
        last_paid_deposit = None

        gen = self._deposit_generator(context, deposit_meta.available_head_id_of_virtual_step)
        for deposit in filter(lambda d: block_height < d.expires, gen):
            available_virtual_step = deposit.remaining_virtual_step

            if required_step < available_virtual_step:
                step = required_step
            else:
                step = available_virtual_step

                # All virtual steps are consumed in this loop.
                # So if this `expires` is the `max expires`, should find the next `max expires`.
                if deposit.expires == deposit_meta.expires_of_virtual_step:
                    should_update_expire = True

            if step > 0:
                deposit.consume_virtual_step(step)
                context.storage.fee.put_deposit(context, deposit)
                last_paid_deposit = deposit

                charged_step += step
                required_step -= step

                if required_step == 0:
                    break

        indices_changed = self._update_virtual_step_indices(
            context, deposit_meta, last_paid_deposit, should_update_expire, block_height)

        return charged_step, indices_changed

    def _update_virtual_step_indices(self,
                                     context: 'IconScoreContext',
                                     deposit_meta: 'DepositMeta',
                                     last_paid_deposit: 'Deposit',
                                     should_update_expire: bool,
                                     block_height: int) -> bool:
        """
        Updates indices of virtual steps to DepositMeta and returns whether there exist changes.
        """
        next_available_deposit = last_paid_deposit

        if last_paid_deposit is not None and last_paid_deposit.remaining_virtual_step == 0:
            # All virtual steps have been consumed in the current deposit
            # so should find the next available virtual steps
            gen = self._deposit_generator(context, last_paid_deposit.next_id)
            next_available_deposit = next(filter(lambda d: block_height < d.expires, gen), None)

        next_available_deposit_id = next_available_deposit.id if next_available_deposit else None
        next_expires = deposit_meta.expires_of_virtual_step

        if next_available_deposit_id is None:
            # This means that there are no available virtual steps in all deposits.
            next_expires = -1
        elif should_update_expire:
            # Finds next max expires. Sets to -1 if not exist.
            gen = self._deposit_generator(context, next_available_deposit_id)
            next_expires = max(map(lambda d: d.expires, gen), default=-1)

        if deposit_meta.available_head_id_of_virtual_step != next_available_deposit_id \
                or deposit_meta.expires_of_virtual_step != next_expires:
            # Updates and return True if some changes exist
            deposit_meta.available_head_id_of_virtual_step = next_available_deposit_id
            deposit_meta.expires_of_virtual_step = next_expires
            return True

        return False

    def _charge_fee_from_deposit(self,
                                 context: 'IconScoreContext',
                                 deposit_meta: 'DepositMeta',
                                 required_icx: int,
                                 block_height: int) -> (int, bool):
        """
        Charges fees from available deposit ICXs
        Returns total charged amount and whether the properties of 'deposit_meta' are changed
        """
        assert required_icx > 0

        if required_icx == 0:
            return 0, False

        remaining_required_icx = required_icx
        should_update_expire = False
        last_paid_deposit = None

        # Search for next available deposit id
        gen = self._deposit_generator(context, deposit_meta.available_head_id_of_deposit)
        for deposit in filter(lambda d: block_height < d.expires, gen):
            available_deposit = deposit.remaining_deposit - deposit.min_remaining_deposit

            if remaining_required_icx < available_deposit:
                charged_icx = remaining_required_icx
            else:
                charged_icx = available_deposit

                # All available deposits are consumed in this loop.
                # So if this `expires` is the `max expires`, should find the next `max expires`.
                if deposit.expires == deposit_meta.expires_of_deposit:
                    should_update_expire = True

            if charged_icx > 0:
                deposit.consume_deposit(charged_icx)
                context.storage.fee.put_deposit(context, deposit)
                last_paid_deposit = deposit

                remaining_required_icx -= charged_icx
                if remaining_required_icx == 0:
                    break

        if remaining_required_icx > 0:
            # Charges all remaining fee regardless of the minimum remaining amount.
            gen = self._deposit_generator(context, deposit_meta.head_id)
            for deposit in filter(lambda d: block_height < d.expires, gen):
                charged_icx = min(remaining_required_icx, deposit.remaining_deposit)

                if charged_icx > 0:
                    deposit.consume_deposit(charged_icx)
                    context.storage.fee.put_deposit(context, deposit)

                    remaining_required_icx -= charged_icx
                    if remaining_required_icx == 0:
                        break

        indices_changed = self._update_deposit_indices(
            context, deposit_meta, last_paid_deposit, should_update_expire, block_height)

        return required_icx - remaining_required_icx, indices_changed

    def _update_deposit_indices(self,
                                context: 'IconScoreContext',
                                deposit_meta: 'DepositMeta',
                                last_paid_deposit: 'Deposit',
                                should_update_expire: bool,
                                block_height: int) -> bool:
        """
        Updates indices of deposit to deposit_meta and returns whether there exist changes.
        """

        next_available_deposit = last_paid_deposit

        if last_paid_deposit.remaining_deposit <= last_paid_deposit.min_remaining_deposit:
            # All available deposits have been consumed in the current deposit
            # so should find the next available deposits
            gen = self._deposit_generator(context, last_paid_deposit.next_id)
            next_available_deposit = next(filter(lambda d: block_height < d.expires, gen), None)

        next_available_deposit_id = next_available_deposit.id if next_available_deposit else None
        next_expires = deposit_meta.expires_of_deposit

        if next_available_deposit_id is None:
            # This means that there are no available deposits.
            next_expires = -1
        elif should_update_expire:
            # Finds next max expires. Sets to -1 if not exist.
            gen = self._deposit_generator(context, next_available_deposit_id)
            next_expires = max(map(lambda d: d.expires, gen), default=-1)

        if deposit_meta.available_head_id_of_deposit != next_available_deposit_id \
                or deposit_meta.expires_of_deposit != next_expires:
            # Updates and return True if changes are exist
            deposit_meta.available_head_id_of_deposit = next_available_deposit_id
            deposit_meta.expires_of_deposit = next_expires
            return True

        return False

    def _deposit_generator(self, context: 'IconScoreContext', start_id: Optional[bytes]):
        next_id = start_id
        while next_id is not None:
            deposit = context.storage.fee.get_deposit(context, next_id)
            if deposit is None:
                break

            yield deposit
            next_id = deposit.next_id

    def _get_score_deploy_info(self, context: 'IconScoreContext', score_address: 'Address') -> 'IconScoreDeployInfo':
        deploy_info: 'IconScoreDeployInfo' = context.storage.deploy.get_deploy_info(context, score_address)

        if deploy_info is None:
            raise InvalidRequestException('Invalid SCORE')

        return deploy_info

    def _check_score_valid(self, context: 'IconScoreContext', score_address: 'Address') -> None:
        deploy_info = self._get_score_deploy_info(context, score_address)
        assert deploy_info is not None

    def _check_score_ownership(self, context: 'IconScoreContext', sender: 'Address', score_address: 'Address') -> None:
        deploy_info = self._get_score_deploy_info(context, score_address)

        if deploy_info.owner != sender:
            raise InvalidRequestException('Invalid SCORE owner')

    @staticmethod
    def _check_deposit_id(deposit_id: bytes) -> None:
        if not (isinstance(deposit_id, bytes) and len(deposit_id) == 32):
            raise InvalidRequestException('Invalid deposit ID')

    def _get_or_create_deposit_meta(
            self, context: 'IconScoreContext', score_address: 'Address') -> 'DepositMeta':
        deposit_meta = context.storage.fee.get_deposit_meta(context, score_address)
        return deposit_meta if deposit_meta else DepositMeta()

    @staticmethod
    def _calculate_penalty(deposit: 'Deposit',
                           block_height: int,
                           step_price: int) -> int:
        assert isinstance(deposit, Deposit)
        assert isinstance(block_height, int)
        assert isinstance(step_price, int)

        if block_height >= deposit.expires:
            return 0

        return VirtualStepCalculator.calculate_penalty(
            deposit.virtual_step_used,
            step_price)


class VirtualStepCalculator:
    """
    Calculator for generating Virtual Step
    """

    @classmethod
    def calculate_virtual_step(cls,
                               deposit_amount: int,
                               term: int,
                               step_price: int) -> int:
        """Returns issuance of virtual-step according to deposit_amount and term

        :param deposit_amount: deposit amount in loop unit
        :param term: deposit term
        :param step_price:
        """
        assert term == BLOCKS_IN_ONE_MONTH
        return int(Decimal(deposit_amount) * Decimal(FIXED_RATIO_PER_MONTH) / Decimal(step_price))

    @classmethod
    def calculate_penalty(cls,
                          virtual_step_used: int,
                          step_price: int) -> int:
        """Returns penalty according to given parameters

        :param virtual_step_used:
        :param step_price:
        """
        return virtual_step_used * step_price


class DepositHandler:
    """
    Deposit Handler
    """

    # For eventlog emitting
    class EventType(IntEnum):
        DEPOSIT = 0
        WITHDRAW = 1

    SIGNATURE_AND_INDEX = (
        # DepositAdded(id: bytes, from_: Address, amount: int, term: int)
        ('DepositAdded(bytes,Address,int,int)', 2),
        # DepositWithdrawn(id: bytes, from_: Address, returnAmount: int, penalty: int)
        ('DepositWithdrawn(bytes,Address,int,int)', 2)
    )

    @staticmethod
    def get_signature_and_index_count(event_type: 'EventType') -> (str, int):
        return DepositHandler.SIGNATURE_AND_INDEX[event_type]

    def __init__(self):
        pass

    def handle_deposit_request(self, context: 'IconScoreContext', data: dict):
        """
        Handles fee request(querying or invoking)

        :param context: IconScoreContext
        :param data: data field
        :return:
        """
        converted_data = TypeConverter.convert(data, ParamType.DEPOSIT_DATA)
        action = converted_data['action']

        try:
            if action == 'add':
                term: int = BLOCKS_IN_ONE_MONTH if FIXED_TERM else converted_data['term']
                self._add_deposit(context, term)
            elif action == 'withdraw':
                self._withdraw_deposit(context, converted_data['id'])
            else:
                raise InvalidRequestException(f"Invalid action: {action}")
        except KeyError:
            # missing required params for the action
            raise InvalidParamsException("Required params not found")

    def _add_deposit(self, context: 'IconScoreContext', term: int):
        context.engine.fee.add_deposit(context, context.tx.hash, context.msg.sender, context.tx.to,
                                       context.msg.value, context.block.height, term)

        event_log_args = [context.tx.hash, context.msg.sender, context.msg.value, term]
        self._emit_event(context, DepositHandler.EventType.DEPOSIT, event_log_args)

    def _withdraw_deposit(self, context: 'IconScoreContext', deposit_id: bytes):
        if context.msg.value != 0:
            raise InvalidRequestException(f'Invalid value: must be zero')
        withdrawal_amount, penalty = context.engine.fee.withdraw_deposit(
            context, context.msg.sender, deposit_id, context.block.height)

        event_log_args = [deposit_id, context.msg.sender, withdrawal_amount, penalty]
        self._emit_event(context, DepositHandler.EventType.WITHDRAW, event_log_args)

    @staticmethod
    def _emit_event(context: 'IconScoreContext', event_type: 'DepositHandler.EventType', event_log_args: list):
        signature, index_count = DepositHandler.get_signature_and_index_count(event_type)

        fee_charge: bool = True if context.revision < Revision.IISS.value else False
        EventLogEmitter.emit_event_log(
            context, context.tx.to, signature, event_log_args, index_count, fee_charge)
